import logging
import os
import shutil
import subprocess
import tempfile
from datetime import datetime
from time import gmtime
from typing import TYPE_CHECKING

from galaxy.tool_shed.util import basic_util
from galaxy.tool_shed.util.hg_util import (
    clone_repository,
    copy_file_from_manifest,
    get_changectx_for_changeset,
    get_config_from_disk,
    get_ctx_file_path_from_manifest,
    get_file_context_from_ctx,
    pull_repository,
    reversed_lower_upper_bounded_changelog,
    reversed_upper_bounded_changelog,
    update_repository,
)
from galaxy.util import unicodify

if TYPE_CHECKING:
    from galaxy.util.path import StrPath

log = logging.getLogger(__name__)

INITIAL_CHANGELOG_HASH = "000000000000"


def add_changeset(repo_path: "StrPath", path_to_filename_in_archive: "StrPath"):
    try:
        subprocess.check_output(["hg", "add", path_to_filename_in_archive], stderr=subprocess.STDOUT, cwd=repo_path)
    except Exception as e:
        error_message = f"Error adding '{path_to_filename_in_archive}' to repository: {unicodify(e)}"
        if isinstance(e, subprocess.CalledProcessError):
            error_message += f"\nOutput was:\n{unicodify(e.output)}"
        raise Exception(error_message)


def archive_repository_revision(app, repository, archive_dir, changeset_revision):
    """Create an un-versioned archive of a repository."""
    repo_path = repository.repo_path(app)
    try:
        subprocess.check_output(
            ["hg", "archive", "-r", changeset_revision, archive_dir], stderr=subprocess.STDOUT, cwd=repo_path
        )
    except Exception as e:
        error_message = f"Error attempting to archive revision '{changeset_revision}' of repository '{repository.name}': {unicodify(e)}"
        if isinstance(e, subprocess.CalledProcessError):
            error_message += f"\nOutput was:\n{unicodify(e.output)}"
        log.exception(error_message)
        raise Exception(error_message)


def commit_changeset(repo_path: "StrPath", full_path_to_changeset: "StrPath", username: str, message: str) -> None:
    try:
        subprocess.check_output(
            ["hg", "commit", "-u", username, "-m", message, full_path_to_changeset],
            stderr=subprocess.STDOUT,
            cwd=repo_path,
        )
    except Exception as e:
        error_message = f"Error committing '{full_path_to_changeset}' to repository: {unicodify(e)}"
        if isinstance(e, subprocess.CalledProcessError):
            if e.returncode == 1 and "nothing changed" in unicodify(e.output):
                return
            error_message += f"\nOutput was:\n{unicodify(e.output)}"
        raise Exception(error_message)


def get_hgrc_path(repo_path):
    return os.path.join(repo_path, ".hg", "hgrc")


def create_hgrc_file(app, repository, repo_path):
    # Since we support both http and https, we set `push_ssl` to False to
    # override the default (which is True) in the Mercurial API.
    # The hg purge extension purges all files and directories not being tracked
    # by Mercurial in the current repository. It will remove unknown files and
    # empty directories. This is not currently used because it is not supported
    # in the Mercurial API.
    hgrc_path = get_hgrc_path(repo_path)
    with open(hgrc_path, "w") as fp:
        fp.write("[paths]\n")
        fp.write("default = .\n")
        fp.write("default-push = .\n")
        fp.write("[web]\n")
        if app.config.config_hg_for_dev:
            allow_push = "*"
        else:
            allow_push = repository.user.username
        fp.write(f"allow_push = {allow_push}\n")
        fp.write(f"name = {repository.name}\n")
        fp.write("push_ssl = false\n")
        fp.write("[extensions]\n")
        fp.write("hgext.purge=")


def get_named_tmpfile_from_ctx(ctx, filename, dir):
    """
    Return a named temporary file created from a specified file with a given name included in a repository
    changeset revision.
    """
    filename = basic_util.strip_path(filename)
    for ctx_file in ctx.files():
        ctx_file_name = basic_util.strip_path(unicodify(ctx_file))
        if filename == ctx_file_name:
            try:
                # If the file was moved, its destination file contents will be returned here.
                fctx = ctx[ctx_file]
            except LookupError:
                # Continue looking in case the file was moved.
                fctx = None
                continue
            if fctx:
                fh = tempfile.NamedTemporaryFile("wb", prefix="tmp-toolshed-gntfc", dir=dir)
                tmp_filename = fh.name
                fh.close()
                fh = open(tmp_filename, "wb")
                fh.write(fctx.data())
                fh.close()
                return tmp_filename
    return None


def get_readable_ctx_date(ctx):
    """Convert the date of the changeset (the received ctx) to a human-readable date."""
    t, tz = ctx.date()
    date = datetime(*gmtime(float(t) - tz)[:6])
    ctx_date = date.strftime("%Y-%m-%d")
    return ctx_date


def get_repository_heads(repo):
    """Return current repository heads, which are changesets with no child changesets."""
    heads = [repo[h] for h in repo.heads(None)]
    return heads


def get_reversed_changelog_changesets(repo):
    """Return a list of changesets in reverse order from that provided by the repository manifest."""
    reversed_changelog = []
    for changeset in repo.changelog:
        reversed_changelog.insert(0, changeset)
    return reversed_changelog


def get_revision_label(app, repository, changeset_revision, include_date=True, include_hash=True):
    """
    Return a string consisting of the human readable changeset rev and the changeset revision string
    which includes the revision date if the receive include_date is True.
    """
    repo = repository.hg_repo
    if ctx := get_changectx_for_changeset(repo, changeset_revision):
        return get_revision_label_from_ctx(ctx, include_date=include_date, include_hash=include_hash)
    else:
        if include_hash:
            return f"-1:{changeset_revision}"
        else:
            return "-1"


def get_rev_label_changeset_revision_from_repository_metadata(
    app, repository_metadata, repository=None, include_date=True, include_hash=True
):
    if repository is None:
        repository = repository_metadata.repository
    repo = repository.hg_repo
    changeset_revision = repository_metadata.changeset_revision
    if ctx := get_changectx_for_changeset(repo, changeset_revision):
        rev = f"{ctx.rev():04d}"
        if include_date:
            changeset_revision_date = get_readable_ctx_date(ctx)
            if include_hash:
                label = f"{ctx.rev()}:{changeset_revision} ({changeset_revision_date})"
            else:
                label = f"{ctx.rev()} ({changeset_revision_date})"
        else:
            if include_hash:
                label = f"{ctx.rev()}:{changeset_revision}"
            else:
                label = f"{ctx.rev()}"
    else:
        rev = "-1"
        if include_hash:
            label = f"-1:{changeset_revision}"
        else:
            label = "-1"
    return rev, label, changeset_revision


def get_revision_label_from_ctx(ctx, include_date=True, include_hash=True):
    if include_date:
        if include_hash:
            return f'{ctx.rev()}:{ctx} <i><font color="#666666">({get_readable_ctx_date(ctx)})</font></i>'
        else:
            return f'{ctx.rev()} <i><font color="#666666">({get_readable_ctx_date(ctx)})</font></i>'
    else:
        if include_hash:
            return f"{ctx.rev()}:{ctx}"
        else:
            return str(ctx.rev())


def get_rev_label_from_changeset_revision(repo, changeset_revision, include_date=True, include_hash=True):
    """
    Given a changeset revision hash, return two strings, the changeset rev and the changeset revision hash
    which includes the revision date if the receive include_date is True.
    """
    if ctx := get_changectx_for_changeset(repo, changeset_revision):
        rev = f"{ctx.rev():04d}"
        label = get_revision_label_from_ctx(ctx, include_date=include_date)
    else:
        rev = "-1"
        label = f"-1:{changeset_revision}"
    return rev, label


def remove_path(repo_path: "StrPath", selected_file: "StrPath"):
    cmd = ["hg", "remove", "--force", selected_file]
    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT, cwd=repo_path)
    except Exception as e:
        error_message = f"Error removing path '{selected_file}': {unicodify(e)}"
        if isinstance(e, subprocess.CalledProcessError):
            output = unicodify(e.output)
            if "is untracked" in output:
                # That's ok, happens if we add a new file or directory via tarball upload,
                # just delete the file or dir on disk
                selected_file_path = os.path.join(repo_path, selected_file)
                if os.path.isdir(selected_file_path):
                    shutil.rmtree(selected_file_path)
                else:
                    os.remove(selected_file_path)
                return
            error_message += f"\nOutput was:\n{output}"
        raise Exception(error_message)


def init_repository(repo_path: "StrPath"):
    """
    Create a new Mercurial repository in the given directory.
    """
    try:
        subprocess.check_output(["hg", "init"], stderr=subprocess.STDOUT, cwd=repo_path)
    except Exception as e:
        error_message = f"Error initializing repository: {unicodify(e)}"
        if isinstance(e, subprocess.CalledProcessError):
            error_message += f"\nOutput was:\n{unicodify(e.output)}"
        raise Exception(error_message)


def changeset2rev(repo_path, changeset_revision):
    """
    Return the revision number (as an int) corresponding to a specified changeset revision.
    """
    try:
        rev = subprocess.check_output(
            ["hg", "id", "-r", changeset_revision, "-n"], stderr=subprocess.STDOUT, cwd=repo_path
        )
    except Exception as e:
        error_message = f"Error looking for changeset '{changeset_revision}': {unicodify(e)}"
        if isinstance(e, subprocess.CalledProcessError):
            error_message += f"\nOutput was:\n{unicodify(e.output)}"
        raise Exception(error_message)
    return int(rev.strip())


__all__ = (
    "add_changeset",
    "archive_repository_revision",
    "clone_repository",
    "commit_changeset",
    "copy_file_from_manifest",
    "create_hgrc_file",
    "get_changectx_for_changeset",
    "get_config_from_disk",
    "get_ctx_file_path_from_manifest",
    "get_file_context_from_ctx",
    "get_named_tmpfile_from_ctx",
    "get_readable_ctx_date",
    "get_repository_heads",
    "get_reversed_changelog_changesets",
    "get_revision_label",
    "get_rev_label_changeset_revision_from_repository_metadata",
    "get_revision_label_from_ctx",
    "get_rev_label_from_changeset_revision",
    "pull_repository",
    "remove_path",
    "reversed_lower_upper_bounded_changelog",
    "reversed_upper_bounded_changelog",
    "update_repository",
    "init_repository",
    "changeset2rev",
)



import logging

from sqlalchemy import (
    and_,
    false,
    or_,
    select,
)
from typing_extensions import Protocol

import tool_shed.repository_types.util as rt_util
from tool_shed.util import (
    hg_util,
    metadata_util,
)
from tool_shed.webapp import model
from tool_shed.webapp.model import (
    Category,
    Repository,
    RepositoryMetadata,
)
from .structured_app import ToolShedApp

log = logging.getLogger(__name__)


class RegistryInterface(Protocol):

    # only used in legacy controllers - can drop when drop old tool shed webapp
    def add_category_entry(self, category): ...

    def add_entry(self, repository):
        # used when creating a repository
        # Used in legacy add_repository_registry_entry API endpoint - already deemed not worth pulling over to 2.0
        # used in legacy undelete_repository admin controller
        # used in legacy deprecate controller
        ...

    def edit_category_entry(self, old_name, new_name):
        # used in legacy admin controller
        ...

    def is_valid(self, repository) -> bool:
        # probably not used outside this class but also not hurting anything
        ...

    def remove_category_entry(self, category): ...

    def remove_entry(self, repository): ...


# stop gap to implement the same general interface as repository_registry but do nothing,
# I don't think this stuff is needed - outside potentially old test cases?
class NullRepositoryRegistry(RegistryInterface):

    def __init__(self, app: ToolShedApp):
        self.app = app

        # all of these are only used by repository_grids - which I think the tool shed 2.0
        # does not use at all - but they are part of the "public interface" consumed by
        # the full registry.
        self.certified_level_one_viewable_repositories_and_suites_by_category: dict = {}
        self.certified_level_one_viewable_suites_by_category: dict = {}
        self.viewable_repositories_and_suites_by_category: dict = {}
        self.viewable_suites_by_category: dict = {}
        self.viewable_valid_repositories_and_suites_by_category: dict = {}
        self.viewable_valid_suites_by_category: dict = {}

    def add_category_entry(self, category):
        # only used in legacy controllers - can drop when drop old tool shed webapp
        pass

    def add_entry(self, repository):
        # used when creating a repository, and maybe more?
        pass

    def edit_category_entry(self, old_name, new_name):
        pass

    def is_valid(self, repository) -> bool:
        if repository and not repository.deleted and not repository.deprecated and repository.downloadable_revisions:
            return True
        return False

    def remove_category_entry(self, category):
        pass

    def remove_entry(self, repository):
        pass


class Registry(RegistryInterface):
    def __init__(self, app):
        log.debug("Loading the repository registry...")
        self.app = app
        # The following lists contain tuples like ( repository.name, repository.user.username, changeset_revision )
        # where the changeset_revision entry is always the latest installable changeset_revision..
        self.certified_level_one_repository_and_suite_tuples = []
        self.certified_level_one_suite_tuples = []
        # These category dictionaries contain entries where the key is the category and the value is the integer count
        # of viewable repositories within that category.

        # only used internally to class and by repository_grids
        self.certified_level_one_viewable_repositories_and_suites_by_category = {}
        # only used internally to class and by repository_grids
        self.certified_level_one_viewable_suites_by_category = {}

        # not even used in legacy shed code...
        # self.certified_level_two_repository_and_suite_tuples = []
        # self.certified_level_two_suite_tuples = []
        # self.certified_level_two_viewable_repositories_and_suites_by_category = {}
        # self.certified_level_two_viewable_suites_by_category = {}

        # only used internally to class
        self.repository_and_suite_tuples = []
        # only used internally to class
        self.suite_tuples = []

        # only used internally to class and by repository_grids
        self.viewable_repositories_and_suites_by_category = {}
        # only used internally to class and by repository_grids
        self.viewable_suites_by_category = {}
        # only used internally to class and by repository_grids
        self.viewable_valid_repositories_and_suites_by_category = {}
        # only used internally to class and by repository_grids
        self.viewable_valid_suites_by_category = {}
        self.load()

    def load(self):
        with self.sa_session.begin():
            self.load_viewable_repositories_and_suites_by_category()
            self.load_repository_and_suite_tuples()

    def add_category_entry(self, category):
        category_name = str(category.name)
        if category_name not in self.viewable_repositories_and_suites_by_category:
            self.viewable_repositories_and_suites_by_category[category_name] = 0
        if category_name not in self.viewable_suites_by_category:
            self.viewable_suites_by_category[category_name] = 0
        if category_name not in self.viewable_valid_repositories_and_suites_by_category:
            self.viewable_valid_repositories_and_suites_by_category[category_name] = 0
        if category_name not in self.viewable_valid_suites_by_category:
            self.viewable_valid_suites_by_category[category_name] = 0
        if category_name not in self.certified_level_one_viewable_repositories_and_suites_by_category:
            self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] = 0
        if category_name not in self.certified_level_one_viewable_suites_by_category:
            self.certified_level_one_viewable_suites_by_category[category_name] = 0

    def add_entry(self, repository):
        try:
            if repository:
                is_valid = self.is_valid(repository)
                certified_level_one_tuple = self.get_certified_level_one_tuple(repository)
                latest_installable_changeset_revision, is_level_one_certified = certified_level_one_tuple
                for rca in repository.categories:
                    category = rca.category
                    category_name = str(category.name)
                    if category_name in self.viewable_repositories_and_suites_by_category:
                        self.viewable_repositories_and_suites_by_category[category_name] += 1
                    else:
                        self.viewable_repositories_and_suites_by_category[category_name] = 1
                    if is_valid:
                        if category_name in self.viewable_valid_repositories_and_suites_by_category:
                            self.viewable_valid_repositories_and_suites_by_category[category_name] += 1
                        else:
                            self.viewable_valid_repositories_and_suites_by_category[category_name] = 1
                    if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                        if category_name in self.viewable_suites_by_category:
                            self.viewable_suites_by_category[category_name] += 1
                        else:
                            self.viewable_suites_by_category[category_name] = 1
                        if is_valid:
                            if category_name in self.viewable_valid_suites_by_category:
                                self.viewable_valid_suites_by_category[category_name] += 1
                            else:
                                self.viewable_valid_suites_by_category[category_name] = 1
                    if is_level_one_certified:
                        if category_name in self.certified_level_one_viewable_repositories_and_suites_by_category:
                            self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] += 1
                        else:
                            self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] = 1
                        if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                            if category_name in self.certified_level_one_viewable_suites_by_category:
                                self.certified_level_one_viewable_suites_by_category[category_name] += 1
                            else:
                                self.certified_level_one_viewable_suites_by_category[category_name] = 1
                self.load_repository_and_suite_tuple(repository)
                if is_level_one_certified:
                    self.load_certified_level_one_repository_and_suite_tuple(repository)
        except Exception:
            # The viewable repository numbers and the categorized (filtered) lists of repository tuples
            # may be slightly skewed, but that is no reason to result in a potential server error.  All
            # will be corrected at next server start.
            log.exception("Handled error adding entry to repository registry")

    def edit_category_entry(self, old_name, new_name):
        if old_name in self.viewable_repositories_and_suites_by_category:
            val = self.viewable_repositories_and_suites_by_category[old_name]
            del self.viewable_repositories_and_suites_by_category[old_name]
            self.viewable_repositories_and_suites_by_category[new_name] = val
        else:
            self.viewable_repositories_and_suites_by_category[new_name] = 0
        if old_name in self.viewable_valid_repositories_and_suites_by_category:
            val = self.viewable_valid_repositories_and_suites_by_category[old_name]
            del self.viewable_valid_repositories_and_suites_by_category[old_name]
            self.viewable_valid_repositories_and_suites_by_category[new_name] = val
        else:
            self.viewable_valid_repositories_and_suites_by_category[new_name] = 0
        if old_name in self.viewable_suites_by_category:
            val = self.viewable_suites_by_category[old_name]
            del self.viewable_suites_by_category[old_name]
            self.viewable_suites_by_category[new_name] = val
        else:
            self.viewable_suites_by_category[new_name] = 0
        if old_name in self.viewable_valid_suites_by_category:
            val = self.viewable_valid_suites_by_category[old_name]
            del self.viewable_valid_suites_by_category[old_name]
            self.viewable_valid_suites_by_category[new_name] = val
        else:
            self.viewable_valid_suites_by_category[new_name] = 0
        if old_name in self.certified_level_one_viewable_repositories_and_suites_by_category:
            val = self.certified_level_one_viewable_repositories_and_suites_by_category[old_name]
            del self.certified_level_one_viewable_repositories_and_suites_by_category[old_name]
            self.certified_level_one_viewable_repositories_and_suites_by_category[new_name] = val
        else:
            self.certified_level_one_viewable_repositories_and_suites_by_category[new_name] = 0
        if old_name in self.certified_level_one_viewable_suites_by_category:
            val = self.certified_level_one_viewable_suites_by_category[old_name]
            del self.certified_level_one_viewable_suites_by_category[old_name]
            self.certified_level_one_viewable_suites_by_category[new_name] = val
        else:
            self.certified_level_one_viewable_suites_by_category[new_name] = 0

    def get_certified_level_one_clause_list(self):
        # only used internally to class
        clause_list = []
        for repository in get_repositories(self.sa_session):
            certified_level_one_tuple = self.get_certified_level_one_tuple(repository)
            latest_installable_changeset_revision, is_level_one_certified = certified_level_one_tuple
            if is_level_one_certified:
                clause_list.append(
                    and_(
                        RepositoryMetadata.repository_id == repository.id,
                        RepositoryMetadata.changeset_revision == latest_installable_changeset_revision,
                    )
                )
        return clause_list

    def get_certified_level_one_tuple(self, repository):
        """
        Return True if the latest installable changeset_revision of the received repository is level one certified.
        """
        # only used internally to class
        if repository is None:
            return (None, False)
        if repository.deleted or repository.deprecated:
            return (None, False)
        # Get the latest installable changeset revision since that is all that is currently configured for testing.
        latest_installable_changeset_revision = metadata_util.get_latest_downloadable_changeset_revision(
            self.app, repository
        )
        if latest_installable_changeset_revision not in [None, hg_util.INITIAL_CHANGELOG_HASH]:
            encoded_repository_id = self.app.security.encode_id(repository.id)
            repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
                self.app, encoded_repository_id, latest_installable_changeset_revision
            )
            if repository_metadata:
                # No repository_metadata.
                return (latest_installable_changeset_revision, True)
        else:
            # No installable changeset_revision.
            return (None, False)

    def is_level_one_certified(self, repository_metadata):
        # only used internally to class
        if repository_metadata:
            repository = repository_metadata.repository
            if repository:
                if repository.deprecated or repository.deleted:
                    return False
                tuple = (
                    str(repository.name),
                    str(repository.user.username),
                    str(repository_metadata.changeset_revision),
                )
                if repository.type in [rt_util.REPOSITORY_SUITE_DEFINITION]:
                    return tuple in self.certified_level_one_suite_tuples
                else:
                    return tuple in self.certified_level_one_repository_and_suite_tuples
        return False

    def is_valid(self, repository) -> bool:
        if repository and not repository.deleted and not repository.deprecated and repository.downloadable_revisions:
            return True
        return False

    def load_certified_level_one_repository_and_suite_tuple(self, repository):
        # only used internally to class
        # The received repository has been determined to be level one certified.
        name = str(repository.name)
        owner = str(repository.user.username)
        tip_changeset_hash = repository.tip()
        if tip_changeset_hash != hg_util.INITIAL_CHANGELOG_HASH:
            certified_level_one_tuple = (name, owner, tip_changeset_hash)
            if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                if certified_level_one_tuple not in self.certified_level_one_suite_tuples:
                    self.certified_level_one_suite_tuples.append(certified_level_one_tuple)
            else:
                if certified_level_one_tuple not in self.certified_level_one_repository_and_suite_tuples:
                    self.certified_level_one_repository_and_suite_tuples.append(certified_level_one_tuple)

    def load_repository_and_suite_tuple(self, repository):
        # only used internally to class
        name = str(repository.name)
        owner = str(repository.user.username)
        for repository_metadata in repository.metadata_revisions:
            changeset_revision = str(repository_metadata.changeset_revision)
            tuple = (name, owner, changeset_revision)
            if tuple not in self.repository_and_suite_tuples:
                self.repository_and_suite_tuples.append(tuple)
            if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                if tuple not in self.suite_tuples:
                    self.suite_tuples.append(tuple)

    def load_repository_and_suite_tuples(self):
        # only used internally to class
        # Load self.certified_level_one_repository_and_suite_tuples and self.certified_level_one_suite_tuples.
        clauses = self.get_certified_level_one_clause_list()
        for repository in get_certified_repositories_with_user(self.sa_session, clauses, model.User):
            self.load_certified_level_one_repository_and_suite_tuple(repository)
        # Load self.repository_and_suite_tuples and self.suite_tuples
        for repository in get_repositories_with_user(self.sa_session, model.User):
            self.load_repository_and_suite_tuple(repository)

    def load_viewable_repositories_and_suites_by_category(self):
        # only used internally to class
        # Clear all dictionaries just in case they were previously loaded.
        self.certified_level_one_viewable_repositories_and_suites_by_category = {}
        self.certified_level_one_viewable_suites_by_category = {}
        # self.certified_level_two_viewable_repositories_and_suites_by_category = {}
        # self.certified_level_two_viewable_suites_by_category = {}
        self.viewable_repositories_and_suites_by_category = {}
        self.viewable_suites_by_category = {}
        self.viewable_valid_repositories_and_suites_by_category = {}
        self.viewable_valid_suites_by_category = {}
        for category in self.sa_session.scalars(select(Category)):
            category_name = str(category.name)
            if category not in self.certified_level_one_viewable_repositories_and_suites_by_category:
                self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] = 0
            if category not in self.certified_level_one_viewable_suites_by_category:
                self.certified_level_one_viewable_suites_by_category[category_name] = 0
            if category not in self.viewable_repositories_and_suites_by_category:
                self.viewable_repositories_and_suites_by_category[category_name] = 0
            if category not in self.viewable_suites_by_category:
                self.viewable_suites_by_category[category_name] = 0
            if category not in self.viewable_valid_repositories_and_suites_by_category:
                self.viewable_valid_repositories_and_suites_by_category[category_name] = 0
            if category not in self.viewable_valid_suites_by_category:
                self.viewable_valid_suites_by_category[category_name] = 0
            for rca in category.repositories:
                repository = rca.repository
                if not repository.deleted and not repository.deprecated:
                    is_valid = self.is_valid(repository)
                    encoded_repository_id = self.app.security.encode_id(repository.id)
                    tip_changeset_hash = repository.tip()
                    repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
                        self.app, encoded_repository_id, tip_changeset_hash
                    )
                    self.viewable_repositories_and_suites_by_category[category_name] += 1
                    if is_valid:
                        self.viewable_valid_repositories_and_suites_by_category[category_name] += 1
                    if repository.type in [rt_util.REPOSITORY_SUITE_DEFINITION]:
                        self.viewable_suites_by_category[category_name] += 1
                        if is_valid:
                            self.viewable_valid_suites_by_category[category_name] += 1
                    if self.is_level_one_certified(repository_metadata):
                        self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] += 1
                        if repository.type in [rt_util.REPOSITORY_SUITE_DEFINITION]:
                            self.certified_level_one_viewable_suites_by_category[category_name] += 1

    def remove_category_entry(self, category):
        catgeory_name = str(category.name)
        if catgeory_name in self.viewable_repositories_and_suites_by_category:
            del self.viewable_repositories_and_suites_by_category[catgeory_name]
        if catgeory_name in self.viewable_valid_repositories_and_suites_by_category:
            del self.viewable_valid_repositories_and_suites_by_category[catgeory_name]
        if catgeory_name in self.viewable_suites_by_category:
            del self.viewable_suites_by_category[catgeory_name]
        if catgeory_name in self.viewable_valid_suites_by_category:
            del self.viewable_valid_suites_by_category[catgeory_name]
        if catgeory_name in self.certified_level_one_viewable_repositories_and_suites_by_category:
            del self.certified_level_one_viewable_repositories_and_suites_by_category[catgeory_name]
        if catgeory_name in self.certified_level_one_viewable_suites_by_category:
            del self.certified_level_one_viewable_suites_by_category[catgeory_name]

    def remove_entry(self, repository):
        try:
            if repository:
                is_valid = self.is_valid(repository)
                certified_level_one_tuple = self.get_certified_level_one_tuple(repository)
                latest_installable_changeset_revision, is_level_one_certified = certified_level_one_tuple
                for rca in repository.categories:
                    category = rca.category
                    category_name = str(category.name)
                    if category_name in self.viewable_repositories_and_suites_by_category:
                        if self.viewable_repositories_and_suites_by_category[category_name] > 0:
                            self.viewable_repositories_and_suites_by_category[category_name] -= 1
                    else:
                        self.viewable_repositories_and_suites_by_category[category_name] = 0
                    if is_valid:
                        if category_name in self.viewable_valid_repositories_and_suites_by_category:
                            if self.viewable_valid_repositories_and_suites_by_category[category_name] > 0:
                                self.viewable_valid_repositories_and_suites_by_category[category_name] -= 1
                        else:
                            self.viewable_valid_repositories_and_suites_by_category[category_name] = 0
                    if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                        if category_name in self.viewable_suites_by_category:
                            if self.viewable_suites_by_category[category_name] > 0:
                                self.viewable_suites_by_category[category_name] -= 1
                        else:
                            self.viewable_suites_by_category[category_name] = 0
                        if is_valid:
                            if category_name in self.viewable_valid_suites_by_category:
                                if self.viewable_valid_suites_by_category[category_name] > 0:
                                    self.viewable_valid_suites_by_category[category_name] -= 1
                            else:
                                self.viewable_valid_suites_by_category[category_name] = 0
                    if is_level_one_certified:
                        if category_name in self.certified_level_one_viewable_repositories_and_suites_by_category:
                            if self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] > 0:
                                self.certified_level_one_viewable_repositories_and_suites_by_category[
                                    category_name
                                ] -= 1
                        else:
                            self.certified_level_one_viewable_repositories_and_suites_by_category[category_name] = 0
                        if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                            if category_name in self.certified_level_one_viewable_suites_by_category:
                                if self.certified_level_one_viewable_suites_by_category[category_name] > 0:
                                    self.certified_level_one_viewable_suites_by_category[category_name] -= 1
                            else:
                                self.certified_level_one_viewable_suites_by_category[category_name] = 0
                self.unload_repository_and_suite_tuple(repository)
                if is_level_one_certified:
                    self.unload_certified_level_one_repository_and_suite_tuple(repository)
        except Exception:
            # The viewable repository numbers and the categorized (filtered) lists of repository tuples
            # may be slightly skewed, but that is no reason to result in a potential server error.  All
            # will be corrected at next server start.
            log.exception("Handled error removing entry from repository registry")

    @property
    def sa_session(self):
        # only used internally to class
        return self.app.model.session

    def unload_certified_level_one_repository_and_suite_tuple(self, repository):
        # The received repository has been determined to be level one certified.
        name = str(repository.name)
        owner = str(repository.user.username)
        tip_changeset_hash = repository.tip()
        if tip_changeset_hash != hg_util.INITIAL_CHANGELOG_HASH:
            certified_level_one_tuple = (name, owner, tip_changeset_hash)
            if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                if certified_level_one_tuple in self.certified_level_one_suite_tuples:
                    self.certified_level_one_suite_tuples.remove(certified_level_one_tuple)
            else:
                if certified_level_one_tuple in self.certified_level_one_repository_and_suite_tuples:
                    self.certified_level_one_repository_and_suite_tuples.remove(certified_level_one_tuple)

    def unload_repository_and_suite_tuple(self, repository):
        name = str(repository.name)
        owner = str(repository.user.username)
        for repository_metadata in repository.metadata_revisions:
            changeset_revision = str(repository_metadata.changeset_revision)
            tuple = (name, owner, changeset_revision)
            if tuple in self.repository_and_suite_tuples:
                self.repository_and_suite_tuples.remove(tuple)
            if repository.type == rt_util.REPOSITORY_SUITE_DEFINITION:
                if tuple in self.suite_tuples:
                    self.suite_tuples.remove(tuple)


def get_repositories(session):
    stmt = select(Repository).where(Repository.deleted == false()).where(Repository.deprecated == false())
    return session.scalars(stmt)


def get_repositories_with_user(session, user_model):
    stmt = (
        select(Repository).where(Repository.deleted == false()).where(Repository.deprecated == false()).join(user_model)
    )
    return session.scalars(stmt)


def get_certified_repositories_with_user(session, where_clauses, user_model):
    stmt = select(Repository).join(RepositoryMetadata).where(or_(*where_clauses)).join(user_model)
    return session.scalars(stmt)



from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

emboss_repository_name = "emboss_0090"
emboss_repository_description = "Galaxy's emboss tool"
emboss_repository_long_description = "Long description of Galaxy's emboss tool"

filtering_repository_name = "filtering_0090"
filtering_repository_description = "Galaxy's filtering tool"
filtering_repository_long_description = "Long description of Galaxy's filtering tool"

freebayes_repository_name = "freebayes_0090"
freebayes_repository_description = "Galaxy's freebayes tool"
freebayes_repository_long_description = "Long description of Galaxy's freebayes tool"

bwa_base_repository_name = "bwa_base_0090"
bwa_base_repository_description = "BWA Base"
bwa_base_repository_long_description = "NT space mapping with BWA"

bwa_color_repository_name = "bwa_color_0090"
bwa_color_repository_description = "BWA Color"
bwa_color_repository_long_description = "Color space mapping with BWA"

category_name = "Test 0090 Tool Search And Installation"
category_description = "Test 0090 Tool Search And Installation"


class TestRepositoryCircularDependenciesAgain(ShedTwillTestCase):
    """Test more features related to repository dependencies."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_bwa_base_repository(self):
        """Create and populate bwa_base_0090."""
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=bwa_base_repository_name,
            description=bwa_base_repository_description,
            long_description=bwa_base_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "bwa/bwa_base.tar",
            commit_message="Uploaded BWA tarball.",
        )

    def test_0010_create_bwa_color_repository(self):
        """Create and populate bwa_color_0090."""
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=bwa_color_repository_name,
            description=bwa_color_repository_description,
            long_description=bwa_color_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "bwa/bwa_color.tar",
            commit_message="Uploaded BWA color tarball.",
        )

    def test_0020_create_emboss_repository(self):
        """Create and populate emboss_0090."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=emboss_repository_name,
            description=emboss_repository_description,
            long_description=emboss_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "emboss/emboss.tar",
            commit_message="Uploaded emboss tarball.",
        )

    def test_0025_create_filtering_repository(self):
        """Create and populate filtering_0090."""
        category = self.create_category(name=category_name, description=category_description)
        filtering_repository = self.get_or_create_repository(
            name=filtering_repository_name,
            description=filtering_repository_description,
            long_description=filtering_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            filtering_repository,
            "filtering/filtering_1.1.0.tar",
            commit_message="Uploaded filtering 1.1.0 tarball.",
        )

    def test_0030_create_freebayes_repository(self):
        """Create and populate freebayes_0090."""
        category = self.create_category(name=category_name, description=category_description)
        repository = self.get_or_create_repository(
            name=freebayes_repository_name,
            description=freebayes_repository_description,
            long_description=freebayes_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.commit_tar_to_repository(
            repository,
            "freebayes/freebayes.tar",
            commit_message="Uploaded freebayes tarball.",
        )

    def test_0035_create_and_upload_dependency_definitions(self):
        """Create and upload repository dependency definitions."""
        bwa_color_repository = self._get_repository_by_name_and_owner(
            bwa_color_repository_name, common.test_user_1_name
        )
        bwa_base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        filtering_repository = self._get_repository_by_name_and_owner(
            filtering_repository_name, common.test_user_1_name
        )
        freebayes_repository = self._get_repository_by_name_and_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        dependency_xml_path = self.generate_temp_path("test_0090", additional_paths=["freebayes"])
        freebayes_tuple = (
            self.url,
            freebayes_repository.name,
            freebayes_repository.owner,
            self.get_repository_tip(freebayes_repository),
        )
        emboss_tuple = (
            self.url,
            emboss_repository.name,
            emboss_repository.owner,
            self.get_repository_tip(emboss_repository),
        )
        filtering_tuple = (
            self.url,
            filtering_repository.name,
            filtering_repository.owner,
            self.get_repository_tip(filtering_repository),
        )
        self.create_repository_dependency(
            repository=filtering_repository, repository_tuples=[freebayes_tuple], filepath=dependency_xml_path
        )
        self.create_repository_dependency(
            repository=bwa_base_repository, repository_tuples=[emboss_tuple], filepath=dependency_xml_path
        )
        self.create_repository_dependency(
            repository=bwa_color_repository, repository_tuples=[filtering_tuple], filepath=dependency_xml_path
        )

    def test_0040_verify_repository_dependencies(self):
        """Verify the generated dependency structure."""
        bwa_color_repository = self._get_repository_by_name_and_owner(
            bwa_color_repository_name, common.test_user_1_name
        )
        bwa_base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
        emboss_repository = self._get_repository_by_name_and_owner(emboss_repository_name, common.test_user_1_name)
        filtering_repository = self._get_repository_by_name_and_owner(
            filtering_repository_name, common.test_user_1_name
        )
        freebayes_repository = self._get_repository_by_name_and_owner(
            freebayes_repository_name, common.test_user_1_name
        )
        self.check_repository_dependency(filtering_repository, freebayes_repository)
        self.check_repository_dependency(bwa_base_repository, emboss_repository)
        self.check_repository_dependency(bwa_color_repository, filtering_repository)



import logging
import os
from typing import (
    cast,
    Iterable,
    Optional,
)

import alembic
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError

from galaxy.model.database_utils import (
    create_database,
    database_exists,
)
from galaxy.model.migrations.base import (
    BaseAlembicManager,
    DatabaseStateCache,
    get_url_string,
    load_metadata,
)
from galaxy.model.migrations.exceptions import (
    IncorrectSAMigrateVersionError,
    NoVersionTableError,
    OutdatedDatabaseError,
    SAMigrateError,
)
from tool_shed.webapp.model import Base

SQLALCHEMYMIGRATE_LAST_VERSION = 27
TOOLSHED = "tool shed"

log = logging.getLogger(__name__)


class RevisionNotFoundError(Exception):
    # The database has an Alembic version table; however, that table is either empty,
    # or it does not contain a revision id that matches a revision in the code base.
    # As a result, it is impossible to determine the state of the database.
    def __init__(self) -> None:
        msg = f"The {TOOLSHED} database has an alembic version table, but that table does not contain "
        msg += "a revision id that matches a revision in the code base."
        super().__init__(msg)


def verify_database(url, engine_options=None) -> None:
    engine_options = engine_options or {}
    engine = create_engine(url, **engine_options)
    verifier = DatabaseStateVerifier(engine)
    verifier.run()
    engine.dispose()


class AlembicManager(BaseAlembicManager):
    def _get_alembic_root(self) -> str:
        return os.path.dirname(__file__)

    def stamp_head(self) -> None:
        self.stamp_revision("head")

    def is_up_to_date(self) -> bool:
        """
        True if the head revision in the script directory is stored in the database.
        """
        head_id = self.get_script_head()
        return bool(self.db_heads and head_id in self.db_heads)

    def is_under_version_control(self) -> bool:
        """
        True if the database version table contains a revision that corresponds to a revision
        in the script directory.
        """
        if self.db_heads:
            for db_head in self.db_heads:
                try:
                    revision = self._get_revision(db_head)
                    if revision:
                        log.info(f"The version in the database is {db_head}.")
                        return True
                except alembic.util.exc.CommandError:  # No need to raise exception.
                    log.info(f"Revision {db_head} does not exist in the script directory.")
        return False

    def get_db_head(self) -> Optional[str]:
        return self._get_head_revision(cast(Iterable[str], self.db_heads))

    def get_script_head(self) -> Optional[str]:
        return self.script_directory.get_current_head()

    def _get_head_revision(self, heads: Iterable[str]) -> Optional[str]:
        for head in heads:
            if self._get_revision(head):
                return head
        return None


class DatabaseStateVerifier:
    def __init__(self, engine):
        self.engine = engine
        self.metadata = Base.metadata
        # These values may or may not be required, so do a lazy load.
        self._db_state: Optional[DatabaseStateCache] = None
        self._alembic_manager: Optional[AlembicManager] = None

    @property
    def db_state(self) -> DatabaseStateCache:
        if not self._db_state:
            self._db_state = DatabaseStateCache(engine=self.engine)
        return self._db_state

    @property
    def alembic_manager(self) -> AlembicManager:
        if not self._alembic_manager:
            self._alembic_manager = AlembicManager(self.engine)
        return self._alembic_manager

    def run(self) -> None:
        if self._handle_no_database():
            return
        if self._handle_empty_database():
            return
        self._handle_nonempty_database()

    def _handle_no_database(self) -> bool:
        url = get_url_string(self.engine)

        try:
            # connect using the database name from the sqlalchemy engine
            exists = database_exists(url, database=self.engine.url.database)
        except OperationalError:
            exists = False

        if not exists:
            self._create_database(url)
            self._initialize_database()
            return True
        return False

    def _handle_empty_database(self) -> bool:
        if self._is_database_empty():
            self._initialize_database()
            return True
        return False

    def _handle_nonempty_database(self) -> None:
        if self._has_alembic_version_table():
            self._handle_with_alembic()
        elif self._has_sqlalchemymigrate_version_table():
            if self._is_last_sqlalchemymigrate_version():
                script = self._upgrade_script()
                raise SAMigrateError(TOOLSHED, script)
            else:
                raise IncorrectSAMigrateVersionError(TOOLSHED, SQLALCHEMYMIGRATE_LAST_VERSION)
        else:
            raise NoVersionTableError(TOOLSHED)

    def _handle_with_alembic(self) -> None:
        am = self.alembic_manager
        if am.is_up_to_date():
            log.info(f"Your {TOOLSHED} database is up-to-date")
            return
        if am.is_under_version_control():
            code_version = am.get_script_head()
            db_version = am.get_db_head()
            script = self._upgrade_script()
            raise OutdatedDatabaseError(TOOLSHED, cast(str, db_version), cast(str, code_version), script)
        else:
            raise RevisionNotFoundError()

    def _create_database(self, url: str) -> None:
        log.info(f"Creating database for URI [{url}]")
        create_database(url)

    def _initialize_database(self) -> None:
        load_metadata(self.metadata, self.engine)
        self.alembic_manager.stamp_head()

    def _is_database_empty(self) -> bool:
        return self.db_state.is_database_empty()

    def _has_alembic_version_table(self) -> bool:
        return self.db_state.has_alembic_version_table()

    def _has_sqlalchemymigrate_version_table(self) -> bool:
        return self.db_state.has_sqlalchemymigrate_version_table()

    def _is_last_sqlalchemymigrate_version(self) -> bool:
        return self.db_state.is_last_sqlalchemymigrate_version(SQLALCHEMYMIGRATE_LAST_VERSION)

    def _upgrade_script(self) -> str:
        return "manage_toolshed_db.sh upgrade"



import logging
import sys
import time
from typing import Optional

from sqlalchemy.orm.scoping import scoped_session

import galaxy.datatypes.registry
import galaxy.tools.data
import tool_shed.repository_registry
import tool_shed.repository_types.registry
import tool_shed.webapp.model
from galaxy import auth
from galaxy.app import (
    HaltableContainer,
    SentryClientMixin,
)
from galaxy.config import configure_logging
from galaxy.managers.api_keys import ApiKeyManager
from galaxy.managers.citations import CitationsManager
from galaxy.managers.dbkeys import GenomeBuilds
from galaxy.managers.users import UserManager
from galaxy.model.base import SharedModelMapping
from galaxy.model.tags import CommunityTagHandler
from galaxy.quota import (
    NoQuotaAgent,
    QuotaAgent,
)
from galaxy.security import idencoding
from galaxy.structured_app import BasicSharedApp
from galaxy.web_stack import application_stack_instance
from tool_shed.grids.repository_grid_filter_manager import RepositoryGridFilterManager
from tool_shed.managers.model_cache import ModelCache
from tool_shed.structured_app import ToolShedApp
from tool_shed.util.hgweb_config import hgweb_config_manager
from tool_shed.webapp.model.migrations import verify_database
from . import config

log = logging.getLogger(__name__)


class UniverseApplication(ToolShedApp, SentryClientMixin, HaltableContainer):
    """Encapsulates the state of a Universe application"""

    def __init__(self, **kwd) -> None:
        super().__init__()
        self[BasicSharedApp] = self
        self[ToolShedApp] = self
        log.debug("python path is: %s", ", ".join(sys.path))
        self.name = "tool_shed"
        # will be overwritten when building WSGI app
        self.is_webapp = False
        # Read the tool_shed.ini configuration file and check for errors.
        self.config = config.Configuration(**kwd)
        self.config.check()
        configure_logging(self.config)
        self.application_stack = application_stack_instance()
        # Initialize the  Galaxy datatypes registry.
        self.datatypes_registry = galaxy.datatypes.registry.Registry()
        self.datatypes_registry.load_datatypes(self.config.root, self.config.datatypes_config)
        # Initialize the Tool Shed repository_types registry.
        self.repository_types_registry = tool_shed.repository_types.registry.Registry()
        # Initialize the RepositoryGridFilterManager.
        self.repository_grid_filter_manager = RepositoryGridFilterManager()
        # Determine the Tool Shed database connection string.
        if self.config.database_connection:
            db_url = self.config.database_connection
        else:
            db_url = f"sqlite:///{self.config.database}?isolation_level=IMMEDIATE"

        # Initialize the Tool Shed database and check for appropriate schema version.
        verify_database(db_url, self.config.database_engine_options)

        # Set up the Tool Shed database engine and ORM.
        from tool_shed.webapp.model import mapping

        model: mapping.ToolShedModelMapping = mapping.init(db_url, self.config.database_engine_options)
        self.model = model
        self.security = idencoding.IdEncodingHelper(id_secret=self.config.id_secret)
        self._register_singleton(idencoding.IdEncodingHelper, self.security)
        self._register_singleton(SharedModelMapping, model)
        self._register_singleton(mapping.ToolShedModelMapping, model)
        self._register_singleton(scoped_session, self.model.context)
        self.model_cache = ModelCache(self.config.model_cache_dir)
        self.user_manager = self._register_singleton(UserManager, UserManager(self, app_type="tool_shed"))
        self.api_keys_manager = self._register_singleton(ApiKeyManager)
        # initialize the Tool Shed tag handler.
        self.tag_handler = CommunityTagHandler(self.model.context)
        # Initialize the Tool Shed tool data tables.  Never pass a configuration file here
        # because the Tool Shed should always have an empty dictionary!
        self.tool_data_tables = galaxy.tools.data.ToolDataTableManager(self.config.tool_data_path)
        self.genome_builds = GenomeBuilds(self)
        self.auth_manager = self._register_singleton(auth.AuthManager, auth.AuthManager(self.config))
        # Citation manager needed to load tools.
        self.citations_manager = self._register_singleton(CitationsManager, CitationsManager(self))
        self.use_tool_dependency_resolution = False
        # Initialize the Tool Shed security agent.
        self.security_agent = model.security_agent
        # The Tool Shed makes no use of a quota, but this attribute is still required.
        self.quota_agent = self._register_singleton(QuotaAgent, NoQuotaAgent())
        # Initialize the baseline Tool Shed statistics component.
        self.shed_counter = model.shed_counter
        # Let the Tool Shed's HgwebConfigManager know where the hgweb.config file is located.
        self.hgweb_config_manager = hgweb_config_manager
        self.hgweb_config_manager.hgweb_config_dir = self.config.hgweb_config_dir
        self.hgweb_config_manager.hgweb_repo_prefix = self.config.hgweb_repo_prefix
        # Initialize the repository registry.
        if config.SHED_API_VERSION != "v2":
            self.repository_registry = tool_shed.repository_registry.Registry(self)
        else:
            self.repository_registry = tool_shed.repository_registry.NullRepositoryRegistry(self)
        # Configure Sentry client if configured
        self.configure_sentry_client()
        #  used for cachebusting -- refactor this into a *SINGLE* UniverseApplication base.
        self.server_starttime = int(time.time())
        log.debug("Tool shed hgweb.config file is: %s", self.hgweb_config_manager.hgweb_config)


# Global instance of the universe app.
app: Optional[ToolShedApp] = None



import logging
import os

from ..base import common
from ..base.api import skip_if_api_v2
from ..base.twilltestcase import ShedTwillTestCase

log = logging.getLogger(__name__)

bwa_base_repository_name = "bwa_base_repository_0100"
bwa_base_repository_description = "BWA Base"
bwa_base_repository_long_description = (
    "BWA tool that depends on bwa 0.5.9, with a complex repository dependency pointing at package_bwa_0_5_9_0100"
)

bwa_package_repository_name = "package_bwa_0_5_9_0100"
bwa_package_repository_description = "BWA Tool"
bwa_package_repository_long_description = "BWA repository with a package tool dependency defined for BWA 0.5.9."

category_name = "Test 0100 Complex Repository Dependencies"
category_description = "Test 0100 Complex Repository Dependencies"


class TestComplexRepositoryDependencies(ShedTwillTestCase):
    """Test features related to complex repository dependencies."""

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username, explicit_logout=True)

    def test_0005_create_bwa_package_repository(self):
        """Create and populate package_bwa_0_5_9_0100."""
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        # Create a repository named package_bwa_0_5_9_0100 owned by user1.
        repository = self.get_or_create_repository(
            name=bwa_package_repository_name,
            description=bwa_package_repository_description,
            long_description=bwa_package_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        self.add_file_to_repository(repository, "bwa/complex/tool_dependencies.xml")
        if not self.is_v2:
            # Visit the manage repository page for package_bwa_0_5_9_0100.
            self.display_manage_repository_page(
                repository, strings_displayed=["Tool dependencies", "will not be", "to this repository"]
            )

    def test_0010_create_bwa_base_repository(self):
        """Create and populate bwa_base_0100."""
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        # Create a repository named bwa_base_repository_0100 owned by user1.
        repository = self.get_or_create_repository(
            name=bwa_base_repository_name,
            description=bwa_base_repository_description,
            long_description=bwa_base_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        # Populate the repository named bwa_base_repository_0100 with a bwa_base tool archive.
        self.commit_tar_to_repository(
            repository,
            "bwa/complex/bwa_base.tar",
            commit_message="Uploaded bwa_base.tar with tool wrapper XML, but without tool dependency XML.",
        )

    def test_0015_generate_complex_repository_dependency_invalid_shed_url(self):
        """Generate and upload a complex repository definition that specifies an invalid tool shed URL."""
        dependency_path = self.generate_temp_path("test_0100", additional_paths=["complex", "invalid"])
        # The repository named bwa_base_repository_0100 is the dependent repository.
        base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
        # The repository named package_bwa_0_5_9_0100 is the required repository.
        tool_repository = self._get_repository_by_name_and_owner(bwa_package_repository_name, common.test_user_1_name)
        url = "http://http://this is not an url!"
        name = "package_bwa_0_5_9_0100"
        owner = "user1"
        changeset_revision = self.get_repository_tip(tool_repository)
        strings_displayed = ["Repository dependencies are currently supported only within the same tool shed"]
        # Populate the dependent repository named bwa_base_repository_0100 with an invalid tool_dependencies.xml file.
        repository_tuple = (url, name, owner, changeset_revision)
        self.create_repository_dependency(
            repository=base_repository,
            filepath=dependency_path,
            repository_tuples=[repository_tuple],
            strings_displayed=strings_displayed,
            complex=True,
            package="bwa",
            version="0.5.9",
        )

    def test_0020_generate_complex_repository_dependency_invalid_repository_name(self):
        """Generate and upload a complex repository definition that specifies an invalid repository name."""
        dependency_path = self.generate_temp_path("test_0100", additional_paths=["complex", "invalid"])
        # The base_repository named bwa_base_repository_0100 is the dependent repository.
        base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
        # The repository named package_bwa_0_5_9_0100 is the required repository.
        tool_repository = self._get_repository_by_name_and_owner(bwa_package_repository_name, common.test_user_1_name)
        url = self.url
        name = "invalid_repository!?"
        owner = "user1"
        changeset_revision = self.get_repository_tip(tool_repository)
        strings_displayed = ["because the name is invalid"]
        # Populate the dependent base_repository named package_bwa_0_5_9_0100 with an invalid tool_dependencies.xml file.
        repository_tuple = (url, name, owner, changeset_revision)
        self.create_repository_dependency(
            repository=base_repository,
            filepath=dependency_path,
            repository_tuples=[repository_tuple],
            strings_displayed=strings_displayed,
            complex=True,
            package="bwa",
            version="0.5.9",
        )

    def test_0025_generate_complex_repository_dependency_invalid_owner_name(self):
        """Generate and upload a complex repository definition that specifies an invalid owner."""
        dependency_path = self.generate_temp_path("test_0100", additional_paths=["complex", "invalid"])
        # The base_repository named bwa_base_repository_0100 is the dependent repository.
        base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
        # The repository named package_bwa_0_5_9_0100 is the required repository.
        tool_repository = self._get_repository_by_name_and_owner(bwa_package_repository_name, common.test_user_1_name)
        url = self.url
        name = "package_bwa_0_5_9_0100"
        owner = "invalid_owner!?"
        changeset_revision = self.get_repository_tip(tool_repository)
        strings_displayed = ["because the owner is invalid."]
        repository_tuple = (url, name, owner, changeset_revision)
        self.create_repository_dependency(
            repository=base_repository,
            filepath=dependency_path,
            repository_tuples=[repository_tuple],
            strings_displayed=strings_displayed,
            complex=True,
            package="bwa",
            version="0.5.9",
        )

    def test_0030_generate_complex_repository_dependency_invalid_changeset_revision(self):
        """Generate and upload a complex repository definition that specifies an invalid changeset revision."""
        dependency_path = self.generate_temp_path("test_0100", additional_paths=["complex", "invalid"])
        # The base_repository named bwa_base_repository_0100 is the dependent repository.
        base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
        # The repository named package_bwa_0_5_9_0100 is the required repository.
        url = self.url
        name = "package_bwa_0_5_9_0100"
        owner = "user1"
        changeset_revision = "1234abcd"
        strings_displayed = ["because the changeset revision is invalid."]
        repository_tuple = (url, name, owner, changeset_revision)
        self.create_repository_dependency(
            repository=base_repository,
            filepath=dependency_path,
            repository_tuples=[repository_tuple],
            strings_displayed=strings_displayed,
            complex=True,
            package="bwa",
            version="0.5.9",
        )

    def test_0035_generate_complex_repository_dependency(self):
        """Generate and upload a valid tool_dependencies.xml file that specifies package_bwa_0_5_9_0100."""
        # The base_repository named bwa_base_repository_0100 is the dependent repository.
        base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
        # The repository named package_bwa_0_5_9_0100 is the required repository.
        tool_repository = self._get_repository_by_name_and_owner(bwa_package_repository_name, common.test_user_1_name)
        dependency_path = self.generate_temp_path("test_0100", additional_paths=["complex"])
        url = self.url
        name = "package_bwa_0_5_9_0100"
        owner = "user1"
        changeset_revision = self.get_repository_tip(tool_repository)
        repository_tuple = (url, name, owner, changeset_revision)
        self.create_repository_dependency(
            repository=base_repository,
            filepath=dependency_path,
            repository_tuples=[repository_tuple],
            complex=True,
            package="bwa",
            version="0.5.9",
        )
        self.check_repository_dependency(base_repository, depends_on_repository=tool_repository)
        if not self.is_v2:
            self.display_manage_repository_page(
                base_repository, strings_displayed=["bwa", "0.5.9", "package", changeset_revision]
            )

    @skip_if_api_v2
    def test_0040_generate_tool_dependency(self):
        """Generate and upload a new tool_dependencies.xml file that specifies an arbitrary file on the filesystem, and verify that bwa_base depends on the new changeset revision."""
        # The base_repository named bwa_base_repository_0100 is the dependent repository.
        base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
        # The repository named package_bwa_0_5_9_0100 is the required repository.
        tool_repository = self._get_repository_by_name_and_owner(bwa_package_repository_name, common.test_user_1_name)
        previous_changeset = self.get_repository_tip(tool_repository)
        old_tool_dependency = self.get_filename(os.path.join("bwa", "complex", "readme", "tool_dependencies.xml"))
        new_tool_dependency_path = self.generate_temp_path("test_1100", additional_paths=["tool_dependency"])
        xml_filename = os.path.abspath(os.path.join(new_tool_dependency_path, "tool_dependencies.xml"))
        # Generate a tool_dependencies.xml file that points to an arbitrary file in the local filesystem.
        open(xml_filename, "w").write(
            open(old_tool_dependency).read().replace("__PATH__", self.get_filename("bwa/complex"))
        )
        self.add_file_to_repository(tool_repository, xml_filename, "tool_dependencies.xml")
        # Verify that the dependency display has been updated as a result of the new tool_dependencies.xml file.
        repository_tip = self.get_repository_tip(tool_repository)
        strings_displayed = ["bwa", "0.5.9", "package"]
        strings_displayed.append(repository_tip)
        strings_not_displayed = [previous_changeset]
        self.display_manage_repository_page(
            tool_repository, strings_displayed=strings_displayed, strings_not_displayed=strings_not_displayed
        )
        # Visit the manage page of the package_bwa_0_5_9_0100 to confirm the valid tool dependency definition.
        self.display_manage_repository_page(
            tool_repository, strings_displayed=strings_displayed, strings_not_displayed=strings_not_displayed
        )
        # Visit the manage page of the bwa_base_repository_0100 to confirm the valid tool dependency definition
        # and the updated changeset revision (updated tip) of the package_bwa_0_5_9_0100 repository is displayed
        # as the required repository revision.  The original revision defined in the previously uploaded
        # tool_dependencies.xml file will be updated.
        self.display_manage_repository_page(
            base_repository, strings_displayed=strings_displayed, strings_not_displayed=strings_not_displayed
        )



import os

from ..base import common
from ..base.twilltestcase import ShedTwillTestCase

bwa_base_repository_name = "bwa_base_repository_0100"
bwa_base_repository_description = "BWA Base"
bwa_base_repository_long_description = (
    "BWA tool that depends on bwa 0.5.9, with a complex repository dependency pointing at package_bwa_0_5_9_0100"
)

bwa_package_repository_name = "package_bwa_0_5_9_0100"
bwa_package_repository_description = "BWA Package"
bwa_package_repository_long_description = (
    "BWA repository with a package tool dependency defined to compile and install BWA 0.5.9."
)

category_name = "Test 0100 Complex Repository Dependencies"
category_description = "Test 0100 Complex Repository Dependencies"
running_standalone = False


class TestInstallingComplexRepositoryDependencies(ShedTwillTestCase):
    """Test features related to installing repositories with complex repository dependencies."""

    requires_galaxy = True

    def test_0000_initiate_users(self):
        """Create necessary user accounts."""
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        self.login(email=common.admin_email, username=common.admin_username)

    def test_0005_create_bwa_package_repository(self):
        """Create and populate package_bwa_0_5_9_0100."""
        global running_standalone
        category = self.create_category(name=category_name, description=category_description)
        self.login(email=common.test_user_1_email, username=common.test_user_1_name)
        repository = self.get_or_create_repository(
            name=bwa_package_repository_name,
            description=bwa_package_repository_description,
            long_description=bwa_package_repository_long_description,
            owner=common.test_user_1_name,
            category=category,
            strings_displayed=[],
        )
        if self.repository_is_new(repository):
            running_standalone = True
            old_tool_dependency = self.get_filename(os.path.join("bwa", "complex", "tool_dependencies.xml"))
            new_tool_dependency_path = self.generate_temp_path("test_1100", additional_paths=["tool_dependency"])
            xml_filename = os.path.abspath(os.path.join(new_tool_dependency_path, "tool_dependencies.xml"))
            open(xml_filename, "w").write(
                open(old_tool_dependency).read().replace("__PATH__", self.get_filename("bwa/complex"))
            )
            self.add_file_to_repository(repository, xml_filename, "tool_dependencies.xml")
            self.display_manage_repository_page(
                repository, strings_displayed=["Tool dependencies", "consider setting its type"]
            )

    def test_0010_create_bwa_base_repository(self):
        """Create and populate bwa_base_0100."""
        if running_standalone:
            category = self.create_category(name=category_name, description=category_description)
            self.login(email=common.test_user_1_email, username=common.test_user_1_name)
            repository = self.get_or_create_repository(
                name=bwa_base_repository_name,
                description=bwa_base_repository_description,
                long_description=bwa_base_repository_long_description,
                owner=common.test_user_1_name,
                category=category,
                strings_displayed=[],
            )
            self.commit_tar_to_repository(
                repository,
                "bwa/complex/bwa_base.tar",
                commit_message="Uploaded bwa_base.tar with tool wrapper XML, but without tool dependency XML.",
            )

    def test_0015_generate_complex_repository_dependency_invalid_shed_url(self):
        """Generate and upload a complex repository definition that specifies an invalid tool shed URL."""
        if running_standalone:
            dependency_path = self.generate_temp_path("test_0100", additional_paths=["complex", "shed"])
            base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
            tool_repository = self._get_repository_by_name_and_owner(
                bwa_package_repository_name, common.test_user_1_name
            )
            url = "http://http://this is not an url!"
            name = tool_repository.name
            owner = tool_repository.owner
            changeset_revision = self.get_repository_tip(tool_repository)
            strings_displayed = ["Repository dependencies are currently supported only within the same tool shed"]
            repository_tuple = (url, name, owner, changeset_revision)
            self.create_repository_dependency(
                repository=base_repository,
                filepath=dependency_path,
                repository_tuples=[repository_tuple],
                strings_displayed=strings_displayed,
                complex=True,
                package="bwa",
                version="0.5.9",
            )

    def test_0020_generate_complex_repository_dependency_invalid_repository_name(self):
        """Generate and upload a complex repository definition that specifies an invalid repository name."""
        if running_standalone:
            dependency_path = self.generate_temp_path("test_0100", additional_paths=["complex", "shed"])
            base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
            tool_repository = self._get_repository_by_name_and_owner(
                bwa_package_repository_name, common.test_user_1_name
            )
            url = self.url
            name = "invalid_repository!?"
            owner = tool_repository.owner
            changeset_revision = self.get_repository_tip(tool_repository)
            strings_displayed = ["because the name is invalid."]
            repository_tuple = (url, name, owner, changeset_revision)
            self.create_repository_dependency(
                repository=base_repository,
                filepath=dependency_path,
                repository_tuples=[repository_tuple],
                strings_displayed=strings_displayed,
                complex=True,
                package="bwa",
                version="0.5.9",
            )

    def test_0025_generate_complex_repository_dependency_invalid_owner_name(self):
        """Generate and upload a complex repository definition that specifies an invalid owner."""
        if running_standalone:
            dependency_path = self.generate_temp_path("test_0100", additional_paths=["complex", "shed"])
            base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
            tool_repository = self._get_repository_by_name_and_owner(
                bwa_package_repository_name, common.test_user_1_name
            )
            url = self.url
            name = tool_repository.name
            owner = "invalid_owner!?"
            changeset_revision = self.get_repository_tip(tool_repository)
            strings_displayed = ["because the owner is invalid."]
            repository_tuple = (url, name, owner, changeset_revision)
            self.create_repository_dependency(
                repository=base_repository,
                filepath=dependency_path,
                repository_tuples=[repository_tuple],
                strings_displayed=strings_displayed,
                complex=True,
                package="bwa",
                version="0.5.9",
            )

    def test_0030_generate_complex_repository_dependency_invalid_changeset_revision(self):
        """Generate and upload a complex repository definition that specifies an invalid changeset revision."""
        if running_standalone:
            dependency_path = self.generate_temp_path("test_0100", additional_paths=["complex", "shed"])
            base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
            tool_repository = self._get_repository_by_name_and_owner(
                bwa_package_repository_name, common.test_user_1_name
            )
            url = self.url
            name = tool_repository.name
            owner = tool_repository.owner
            changeset_revision = "1234abcd"
            strings_displayed = ["because the changeset revision is invalid."]
            repository_tuple = (url, name, owner, changeset_revision)
            self.create_repository_dependency(
                repository=base_repository,
                filepath=dependency_path,
                repository_tuples=[repository_tuple],
                strings_displayed=strings_displayed,
                complex=True,
                package="bwa",
                version="0.5.9",
            )

    def test_0035_generate_valid_complex_repository_dependency(self):
        """Generate and upload a valid tool_dependencies.xml file that specifies package_bwa_0_5_9_0100."""
        if running_standalone:
            base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
            tool_repository = self._get_repository_by_name_and_owner(
                bwa_package_repository_name, common.test_user_1_name
            )
            dependency_path = self.generate_temp_path("test_0100", additional_paths=["complex"])
            url = self.url
            name = tool_repository.name
            owner = tool_repository.owner
            changeset_revision = self.get_repository_tip(tool_repository)
            repository_tuple = (url, name, owner, changeset_revision)
            self.create_repository_dependency(
                repository=base_repository,
                filepath=dependency_path,
                repository_tuples=[repository_tuple],
                complex=True,
                package="bwa",
                version="0.5.9",
            )
            self.check_repository_dependency(base_repository, tool_repository)
            self.display_manage_repository_page(base_repository, strings_displayed=["bwa", "0.5.9", "package"])

    def test_0040_update_tool_repository(self):
        """Upload a new tool_dependencies.xml to the tool repository, and verify that the base repository displays the new changeset."""
        if running_standalone:
            base_repository = self._get_repository_by_name_and_owner(bwa_base_repository_name, common.test_user_1_name)
            tool_repository = self._get_repository_by_name_and_owner(
                bwa_package_repository_name, common.test_user_1_name
            )
            previous_changeset = self.get_repository_tip(tool_repository)
            old_tool_dependency = self.get_filename(os.path.join("bwa", "complex", "readme", "tool_dependencies.xml"))
            new_tool_dependency_path = self.generate_temp_path("test_1100", additional_paths=["tool_dependency"])
            xml_filename = os.path.abspath(os.path.join(new_tool_dependency_path, "tool_dependencies.xml"))
            open(xml_filename, "w").write(
                open(old_tool_dependency).read().replace("__PATH__", self.get_filename("bwa/complex"))
            )
            self.add_file_to_repository(tool_repository, xml_filename, "tool_dependencies.xml")
            # Verify that the dependency display has been updated as a result of the new tool_dependencies.xml file.
            self.display_manage_repository_page(
                base_repository,
                strings_displayed=[self.get_repository_tip(tool_repository), "bwa", "0.5.9", "package"],
                strings_not_displayed=[previous_changeset],
            )

    def test_0045_install_base_repository(self):
        """Verify installation of the repository with complex repository dependencies."""
        tool_repository = self._get_repository_by_name_and_owner(bwa_package_repository_name, common.test_user_1_name)
        preview_strings_displayed = [tool_repository.name, self.get_repository_tip(tool_repository)]
        self._install_repository(
            bwa_base_repository_name,
            common.test_user_1_name,
            category_name,
            preview_strings_displayed=preview_strings_displayed,
        )

    def test_0050_verify_installed_repositories(self):
        """Verify that the installed repositories are displayed properly."""
        base_repository = self._get_installed_repository_by_name_owner(
            bwa_base_repository_name, common.test_user_1_name
        )
        tool_repository = self._get_installed_repository_by_name_owner(
            bwa_package_repository_name, common.test_user_1_name
        )
        assert self._get_installed_repository_for(
            common.test_user_1, "bwa_base_repository_0100", base_repository.installed_changeset_revision
        )
        assert self._get_installed_repository_for(
            common.test_user_1, "package_bwa_0_5_9_0100", tool_repository.installed_changeset_revision
        )
        self._assert_has_installed_repository_dependency(base_repository, "package_bwa_0_5_9_0100")



import logging

from galaxy.model.item_attrs import UsesItemRatings

log = logging.getLogger(__name__)


class ItemRatings(UsesItemRatings):
    """Overrides rate_item method since we also allow for comments"""

    def rate_item(self, trans, user, item, rating, comment=""):
        """Rate an item. Return type is <item_class>RatingAssociation."""
        item_rating = self.get_user_item_rating(trans.sa_session, user, item, webapp_model=trans.model)
        if not item_rating:
            # User has not yet rated item; create rating.
            item_rating_assoc_class = self._get_item_rating_assoc_class(item, webapp_model=trans.model)
            item_rating = item_rating_assoc_class()
            item_rating.user = trans.user
            item_rating.set_item(item)
            item_rating.rating = rating
            item_rating.comment = comment
            trans.sa_session.add(item_rating)
            trans.sa_session.commit()
        elif item_rating.rating != rating or item_rating.comment != comment:
            # User has previously rated item; update rating.
            item_rating.rating = rating
            item_rating.comment = comment
            trans.sa_session.add(item_rating)
            trans.sa_session.commit()
        return item_rating



import abc
from typing import (
    List,
    Union,
)

FormValueType = Union[str, bool]


class ShedBrowser(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def visit_url(self, url: str, allowed_codes: List[int]) -> str:
        """Navigate to the supplied URL."""

    @abc.abstractmethod
    def page_content(self) -> str:
        """Return the page content for"""

    @abc.abstractmethod
    def check_page_for_string(self, patt: str) -> None:
        """Looks for 'patt' in the current browser page"""

    @abc.abstractmethod
    def check_string_not_in_page(self, patt: str) -> None:
        """Looks for 'patt' not being in the current browser page"""

    @abc.abstractmethod
    def fill_form_value(self, form_name: str, control_name: str, value: FormValueType):
        """Fill in a form value."""

    @abc.abstractmethod
    def submit_form(self, form_no=-1, button="runtool_btn", form=None, **kwd):
        """Submit the target button."""

    @abc.abstractmethod
    def submit_form_with_name(self, form_name: str, button="runtool_btn", **kwd):
        """Submit the target button."""

    @property
    @abc.abstractmethod
    def is_twill(self) -> bool:
        """Return whether this is a twill browser."""

    @abc.abstractmethod
    def edit_repository_categories(self, categories_to_add: List[str], categories_to_remove: List[str]) -> None:
        """Select some new categories and then restore the component."""

    @abc.abstractmethod
    def grant_users_access(self, usernames: List[str]) -> None:
        """Select users to grant access to."""



import logging

from markupsafe import escape
from sqlalchemy import and_

from galaxy.web.framework.helpers import time_ago
from galaxy.web.legacy_framework import grids
from tool_shed.grids.repository_grids import (
    CategoryGrid,
    RepositoryGrid,
)
from tool_shed.util import hg_util
from tool_shed.webapp import model

log = logging.getLogger(__name__)


class UserGrid(grids.Grid):
    class UserLoginColumn(grids.TextColumn):
        def get_value(self, trans, grid, user):
            return escape(user.email)

    class UserNameColumn(grids.TextColumn):
        def get_value(self, trans, grid, user):
            if user.username:
                return escape(user.username)
            return "not set"

    class GroupsColumn(grids.GridColumn):
        def get_value(self, trans, grid, user):
            if user.groups:
                return len(user.groups)
            return 0

    class RolesColumn(grids.GridColumn):
        def get_value(self, trans, grid, user):
            if user.roles:
                return len(user.roles)
            return 0

    class ExternalColumn(grids.GridColumn):
        def get_value(self, trans, grid, user):
            if user.external:
                return "yes"
            return "no"

    class LastLoginColumn(grids.GridColumn):
        def get_value(self, trans, grid, user):
            if user.current_galaxy_session:
                return self.format(user.current_galaxy_session.update_time)
            return "never"

    class StatusColumn(grids.GridColumn):
        def get_value(self, trans, grid, user):
            if user.purged:
                return "purged"
            elif user.deleted:
                return "deleted"
            return ""

    class EmailColumn(grids.GridColumn):
        def filter(self, trans, user, query, column_filter):
            if column_filter == "All":
                return query
            return query.filter(
                and_(model.Tool.table.c.user_id == model.User.table.c.id, model.User.table.c.email == column_filter)
            )

    title = "Users"
    model_class = model.User
    default_sort_key = "email"
    columns = [
        UserLoginColumn(
            "Email",
            key="email",
            link=(lambda item: dict(operation="information", id=item.id)),
            attach_popup=True,
            filterable="advanced",
        ),
        UserNameColumn("User Name", key="username", attach_popup=False, filterable="advanced"),
        GroupsColumn("Groups", attach_popup=False),
        RolesColumn("Roles", attach_popup=False),
        ExternalColumn("External", attach_popup=False),
        LastLoginColumn("Last Login", format=time_ago),
        StatusColumn("Status", attach_popup=False),
        # Columns that are valid for filtering but are not visible.
        EmailColumn("Email", key="email", visible=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search",
            cols_to_filter=[columns[0], columns[1]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )
    global_actions = [grids.GridAction("Create new user", dict(controller="admin", action="users", operation="create"))]
    operations = [
        grids.GridOperation(
            "Manage Roles and Groups",
            condition=(lambda item: not item.deleted),
            allow_multiple=False,
            url_args=dict(action="manage_roles_and_groups_for_user"),
        ),
        grids.GridOperation(
            "Reset Password",
            condition=(lambda item: not item.deleted),
            allow_multiple=True,
            allow_popup=False,
            url_args=dict(action="reset_user_password"),
        ),
    ]
    standard_filters = [
        grids.GridColumnFilter("Active", args=dict(deleted=False)),
        grids.GridColumnFilter("Deleted", args=dict(deleted=True, purged=False)),
        grids.GridColumnFilter("Purged", args=dict(purged=True)),
        grids.GridColumnFilter("All", args=dict(deleted="All")),
    ]

    use_paging = False

    def get_current_item(self, trans, **kwargs):
        return trans.user


class RoleGrid(grids.Grid):
    class NameColumn(grids.TextColumn):
        def get_value(self, trans, grid, role):
            return escape(str(role.name))

    class DescriptionColumn(grids.TextColumn):
        def get_value(self, trans, grid, role):
            if role.description:
                return str(role.description)
            return ""

    class TypeColumn(grids.TextColumn):
        def get_value(self, trans, grid, role):
            return str(role.type)

    class StatusColumn(grids.GridColumn):
        def get_value(self, trans, grid, role):
            if role.deleted:
                return "deleted"
            return ""

    class GroupsColumn(grids.GridColumn):
        def get_value(self, trans, grid, role):
            if role.groups:
                return len(role.groups)
            return 0

    class RepositoriesColumn(grids.GridColumn):
        def get_value(self, trans, grid, role):
            if role.repositories:
                return len(role.repositories)
            return 0

    class UsersColumn(grids.GridColumn):
        def get_value(self, trans, grid, role):
            if role.users:
                return len(role.users)
            return 0

    title = "Roles"
    model_class = model.Role
    default_sort_key = "name"
    columns = [
        NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="Manage role associations", id=item.id)),
            attach_popup=True,
            filterable="advanced",
        ),
        DescriptionColumn("Description", key="description", attach_popup=False, filterable="advanced"),
        GroupsColumn("Groups", attach_popup=False),
        RepositoriesColumn("Repositories", attach_popup=False),
        UsersColumn("Users", attach_popup=False),
        # Columns that are valid for filtering but are not visible.
        grids.DeletedColumn("Deleted", key="deleted", visible=False, filterable="advanced"),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search", cols_to_filter=[columns[0]], key="free-text-search", visible=False, filterable="standard"
        )
    )
    global_actions = [grids.GridAction("Add new role", dict(controller="admin", action="roles", operation="create"))]
    # Repository admin roles currently do not have any operations since they are managed automatically based
    # on other events.  For example, if a repository is renamed, its associated admin role is automatically
    # renamed accordingly and if a repository is deleted its associated admin role is automatically deleted.
    operations = [
        grids.GridOperation(
            "Rename",
            condition=(lambda item: not item.deleted and not item.is_repository_admin_role),
            allow_multiple=False,
            url_args=dict(action="rename_role"),
        ),
        grids.GridOperation(
            "Delete",
            condition=(lambda item: not item.deleted and not item.is_repository_admin_role),
            allow_multiple=True,
            url_args=dict(action="mark_role_deleted"),
        ),
        grids.GridOperation(
            "Undelete",
            condition=(lambda item: item.deleted and not item.is_repository_admin_role),
            allow_multiple=True,
            url_args=dict(action="undelete_role"),
        ),
        grids.GridOperation(
            "Purge",
            condition=(lambda item: item.deleted and not item.is_repository_admin_role),
            allow_multiple=True,
            url_args=dict(action="purge_role"),
        ),
    ]
    standard_filters = [
        grids.GridColumnFilter("Active", args=dict(deleted=False)),
        grids.GridColumnFilter("Deleted", args=dict(deleted=True)),
        grids.GridColumnFilter("All", args=dict(deleted="All")),
    ]

    use_paging = False

    def apply_query_filter(self, trans, query, **kwd):
        return query.filter(model.Role.type != model.Role.types.PRIVATE)


class GroupGrid(grids.Grid):
    class NameColumn(grids.TextColumn):
        def get_value(self, trans, grid, group):
            return str(group.name)

    class StatusColumn(grids.GridColumn):
        def get_value(self, trans, grid, group):
            if group.deleted:
                return "deleted"
            return ""

    class RolesColumn(grids.GridColumn):
        def get_value(self, trans, grid, group):
            if group.roles:
                return len(group.roles)
            return 0

    class UsersColumn(grids.GridColumn):
        def get_value(self, trans, grid, group):
            if group.users:
                return len(group.users)
            return 0

    title = "Groups"
    model_class = model.Group
    default_sort_key = "name"
    columns = [
        NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="Manage users and roles", id=item.id)),
            attach_popup=True,
        ),
        UsersColumn("Users", attach_popup=False),
        RolesColumn("Roles", attach_popup=False),
        StatusColumn("Status", attach_popup=False),
        # Columns that are valid for filtering but are not visible.
        grids.DeletedColumn("Deleted", key="deleted", visible=False, filterable="advanced"),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search", cols_to_filter=[columns[0]], key="free-text-search", visible=False, filterable="standard"
        )
    )
    global_actions = [grids.GridAction("Add new group", dict(controller="admin", action="groups", operation="create"))]
    operations = [
        grids.GridOperation(
            "Rename",
            condition=(lambda item: not item.deleted),
            allow_multiple=False,
            url_args=dict(action="rename_group"),
        ),
        grids.GridOperation(
            "Delete",
            condition=(lambda item: not item.deleted),
            allow_multiple=True,
            url_args=dict(action="mark_group_deleted"),
        ),
        grids.GridOperation(
            "Undelete",
            condition=(lambda item: item.deleted),
            allow_multiple=True,
            url_args=dict(action="undelete_group"),
        ),
        grids.GridOperation(
            "Purge", condition=(lambda item: item.deleted), allow_multiple=True, url_args=dict(action="purge_group")
        ),
    ]
    standard_filters = [
        grids.GridColumnFilter("Active", args=dict(deleted=False)),
        grids.GridColumnFilter("Deleted", args=dict(deleted=True)),
        grids.GridColumnFilter("All", args=dict(deleted="All")),
    ]

    use_paging = False


class ManageCategoryGrid(CategoryGrid):
    columns = list(CategoryGrid.columns)
    # Override the NameColumn to include an Edit link
    columns[0] = CategoryGrid.NameColumn(
        "Name",
        key="Category.name",
        link=(lambda item: dict(operation="Edit", id=item.id)),
        model_class=model.Category,
        attach_popup=False,
    )
    global_actions = [
        grids.GridAction("Add new category", dict(controller="admin", action="manage_categories", operation="create"))
    ]


class AdminRepositoryGrid(RepositoryGrid):
    class DeletedColumn(grids.BooleanColumn):
        def get_value(self, trans, grid, repository):
            if repository.deleted:
                return "yes"
            return ""

    columns = [
        RepositoryGrid.NameColumn(
            "Name",
            key="name",
            link=(lambda item: dict(operation="view_or_manage_repository", id=item.id)),
            attach_popup=True,
        ),
        RepositoryGrid.HeadsColumn("Heads"),
        RepositoryGrid.UserColumn(
            "Owner",
            model_class=model.User,
            link=(lambda item: dict(operation="repositories_by_user", id=item.id)),
            attach_popup=False,
            key="User.username",
        ),
        RepositoryGrid.DeprecatedColumn("Deprecated", key="deprecated", attach_popup=False),
        # Columns that are valid for filtering but are not visible.
        DeletedColumn("Deleted", key="deleted", attach_popup=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name",
            cols_to_filter=[columns[0]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )
    operations = list(RepositoryGrid.operations)
    operations.append(
        grids.GridOperation(
            "Delete", allow_multiple=False, condition=(lambda item: not item.deleted), async_compatible=False
        )
    )
    operations.append(
        grids.GridOperation(
            "Undelete", allow_multiple=False, condition=(lambda item: item.deleted), async_compatible=False
        )
    )

    def build_initial_query(self, trans, **kwd):
        return trans.sa_session.query(model.Repository).join(model.User.table)


class RepositoryMetadataGrid(grids.Grid):
    class IdColumn(grids.IntegerColumn):
        def get_value(self, trans, grid, repository_metadata):
            return repository_metadata.id

    class NameColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            return escape(repository_metadata.repository.name)

    class OwnerColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            return escape(repository_metadata.repository.user.username)

    class RevisionColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            repository = repository_metadata.repository
            return hg_util.get_revision_label(
                trans.app, repository, repository_metadata.changeset_revision, include_date=True, include_hash=True
            )

    class ToolsColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            tools_str = "0"
            if repository_metadata:
                metadata = repository_metadata.metadata
                if metadata:
                    if "tools" in metadata:
                        # We used to display the following, but grid was too cluttered.
                        # for tool_metadata_dict in metadata[ 'tools' ]:
                        #    tools_str += '%s <b>%s</b><br/>' % ( tool_metadata_dict[ 'id' ], tool_metadata_dict[ 'version' ] )
                        return "{}".format(len(metadata["tools"]))
            return tools_str

    class DatatypesColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            datatypes_str = "0"
            if repository_metadata:
                metadata = repository_metadata.metadata
                if metadata:
                    if "datatypes" in metadata:
                        # We used to display the following, but grid was too cluttered.
                        # for datatype_metadata_dict in metadata[ 'datatypes' ]:
                        #    datatypes_str += '%s<br/>' % datatype_metadata_dict[ 'extension' ]
                        return "{}".format(len(metadata["datatypes"]))
            return datatypes_str

    class WorkflowsColumn(grids.TextColumn):
        def get_value(self, trans, grid, repository_metadata):
            workflows_str = "0"
            if repository_metadata:
                metadata = repository_metadata.metadata
                if metadata:
                    if "workflows" in metadata:
                        # We used to display the following, but grid was too cluttered.
                        # workflows_str += '<b>Workflows:</b><br/>'
                        # metadata[ 'workflows' ] is a list of tuples where each contained tuple is
                        # [ <relative path to the .ga file in the repository>, <exported workflow dict> ]
                        # workflow_tups = metadata[ 'workflows' ]
                        # workflow_metadata_dicts = [ workflow_tup[1] for workflow_tup in workflow_tups ]
                        # for workflow_metadata_dict in workflow_metadata_dicts:
                        #    workflows_str += '%s<br/>' % workflow_metadata_dict[ 'name' ]
                        return "{}".format(len(metadata["workflows"]))
            return workflows_str

    class DeletedColumn(grids.BooleanColumn):
        def get_value(self, trans, grid, repository_metadata):
            if repository_metadata.repository.deleted:
                return "yes"
            return ""

    class DeprecatedColumn(grids.BooleanColumn):
        def get_value(self, trans, grid, repository_metadata):
            if repository_metadata.repository.deprecated:
                return "yes"
            return ""

    class MaliciousColumn(grids.BooleanColumn):
        def get_value(self, trans, grid, repository_metadata):
            if repository_metadata.malicious:
                return "yes"
            return ""

    # Grid definition
    title = "Repository Metadata"
    model_class = model.RepositoryMetadata
    default_sort_key = "name"
    use_hide_message = False
    columns = [
        IdColumn("Id", visible=False, attach_popup=False),
        NameColumn(
            "Name",
            key="name",
            model_class=model.Repository,
            link=(lambda item: dict(operation="view_or_manage_repository_revision", id=item.id)),
            attach_popup=True,
        ),
        OwnerColumn("Owner", attach_popup=False),
        RevisionColumn("Revision", attach_popup=False),
        ToolsColumn("Tools", attach_popup=False),
        DatatypesColumn("Datatypes", attach_popup=False),
        WorkflowsColumn("Workflows", attach_popup=False),
        DeletedColumn("Deleted", attach_popup=False),
        DeprecatedColumn("Deprecated", attach_popup=False),
        MaliciousColumn("Malicious", attach_popup=False),
    ]
    columns.append(
        grids.MulticolFilterColumn(
            "Search repository name",
            cols_to_filter=[columns[1]],
            key="free-text-search",
            visible=False,
            filterable="standard",
        )
    )
    operations = [
        grids.GridOperation(
            "Delete",
            allow_multiple=False,
            allow_popup=True,
            async_compatible=False,
            confirm="Repository metadata records cannot be recovered after they are deleted. Click OK to delete the selected items.",
        )
    ]

    def build_initial_query(self, trans, **kwd):
        return trans.sa_session.query(model.RepositoryMetadata).join(model.Repository.table)



import logging
from typing import (
    Callable,
    Dict,
)

from sqlalchemy import select

from galaxy import (
    util,
    web,
)
from galaxy.webapps.base.controller import HTTPBadRequest
from tool_shed.util import metadata_util
from tool_shed.webapp.model import RepositoryMetadata
from tool_shed.webapp.model.db import get_repository_by_name_and_owner
from . import BaseShedAPIController

log = logging.getLogger(__name__)


class RepositoryRevisionsController(BaseShedAPIController):
    """RESTful controller for interactions with tool shed repository revisions."""

    def __get_value_mapper(self, trans) -> Dict[str, Callable]:
        value_mapper = {
            "id": trans.security.encode_id,
            "repository_id": trans.security.encode_id,
            "user_id": trans.security.encode_id,
        }
        return value_mapper

    @web.legacy_expose_api_anonymous
    def index(self, trans, **kwd):
        """
        GET /api/repository_revisions
        Displays a collection (list) of repository revisions.
        """
        # Example URL: http://localhost:9009/api/repository_revisions
        downloadable = kwd.get("downloadable", None)
        malicious = kwd.get("malicious", None)
        missing_test_components = kwd.get("missing_test_components", None)
        includes_tools = kwd.get("includes_tools", None)
        repository_metadata_dicts = []
        all_repository_metadata = get_repository_metadata(
            trans.sa_session, downloadable, malicious, missing_test_components, includes_tools
        )
        for repository_metadata in all_repository_metadata:
            repository_metadata_dict = repository_metadata.to_dict(
                view="collection", value_mapper=self.__get_value_mapper(trans)
            )
            repository_metadata_dict["url"] = web.url_for(
                controller="repository_revisions", action="show", id=trans.security.encode_id(repository_metadata.id)
            )
            repository_metadata_dicts.append(repository_metadata_dict)
        return repository_metadata_dicts

    @web.legacy_expose_api_anonymous
    def repository_dependencies(self, trans, id, **kwd):
        """
        GET /api/repository_revisions/{encoded repository_metadata id}/repository_dependencies

        Returns a list of dictionaries that each define a specific downloadable revision of a
        repository in the Tool Shed.  This method returns dictionaries with more information in
        them than other methods in this controller.  The information about repository_metdata is
        enhanced to include information about the repository (e.g., name, owner, etc) associated
        with the repository_metadata record.

        :param id: the encoded id of the `RepositoryMetadata` object
        """
        # Example URL: http://localhost:9009/api/repository_revisions/repository_dependencies/bb125606ff9ea620
        repository_dependencies_dicts = []
        repository_metadata = metadata_util.get_repository_metadata_by_id(trans.app, id)
        if repository_metadata is None:
            log.debug(f"Invalid repository_metadata id received: {id}")
            return repository_dependencies_dicts
        metadata = repository_metadata.metadata
        if metadata is None:
            log.debug(f"The repository_metadata record with id {id} has no metadata.")
            return repository_dependencies_dicts
        if "repository_dependencies" in metadata:
            rd_tups = metadata["repository_dependencies"]["repository_dependencies"]
            for rd_tup in rd_tups:
                tool_shed, name, owner, changeset_revision = rd_tup[0:4]
                repository_dependency = get_repository_by_name_and_owner(trans.sa_session, name, owner)
                if repository_dependency is None:
                    log.debug(f"Cannot locate repository dependency {name} owned by {owner}.")
                    continue
                repository_dependency_id = trans.security.encode_id(repository_dependency.id)
                repository_dependency_repository_metadata = metadata_util.get_repository_metadata_by_changeset_revision(
                    trans.app, repository_dependency_id, changeset_revision
                )
                if repository_dependency_repository_metadata is None:
                    # The changeset_revision column in the repository_metadata table has been updated with a new
                    # value value, so find the changeset_revision to which we need to update.
                    new_changeset_revision = metadata_util.get_next_downloadable_changeset_revision(
                        trans.app, repository_dependency, changeset_revision
                    )
                    if new_changeset_revision != changeset_revision:
                        repository_dependency_repository_metadata = (
                            metadata_util.get_repository_metadata_by_changeset_revision(
                                trans.app, repository_dependency_id, new_changeset_revision
                            )
                        )
                        changeset_revision = new_changeset_revision
                    else:
                        decoded_repository_dependency_id = trans.security.decode_id(repository_dependency_id)
                        debug_msg = (
                            f"Cannot locate repository_metadata with id {decoded_repository_dependency_id} for repository dependency {name} owned by {owner} "
                            f"using either of these changeset_revisions: {changeset_revision}, {new_changeset_revision}."
                        )
                        log.debug(debug_msg)
                        continue
                repository_dependency_metadata_dict = repository_dependency_repository_metadata.to_dict(
                    view="element", value_mapper=self.__get_value_mapper(trans)
                )
                repository_dependency_dict = repository_dependency.to_dict(
                    view="element", value_mapper=self.__get_value_mapper(trans)
                )
                # We need to be careful with the entries in our repository_dependency_dict here since this Tool Shed API
                # controller is working with repository_metadata records.  The above to_dict() method returns a dictionary
                # with an id entry for the repository record.  However, all of the other methods in this controller have
                # the id entry associated with a repository_metadata record id.  To avoid confusion, we'll update the
                # repository_dependency_metadata_dict with entries from the repository_dependency_dict without using the
                # Python dictionary update() method because we do not want to overwrite existing entries.
                for k, v in repository_dependency_dict.items():
                    if k not in repository_dependency_metadata_dict:
                        repository_dependency_metadata_dict[k] = v
                repository_dependency_metadata_dict["url"] = web.url_for(
                    controller="repositories", action="show", id=repository_dependency_id
                )
                repository_dependencies_dicts.append(repository_dependency_metadata_dict)
        return repository_dependencies_dicts

    @web.legacy_expose_api_anonymous
    def show(self, trans, id, **kwd):
        """
        GET /api/repository_revisions/{encoded_repository_metadata_id}
        Displays information about a repository_metadata record in the Tool Shed.

        :param id: the encoded id of the `RepositoryMetadata` object
        """
        # Example URL: http://localhost:9009/api/repository_revisions/bb125606ff9ea620
        repository_metadata = metadata_util.get_repository_metadata_by_id(trans.app, id)
        if repository_metadata is None:
            log.debug(f"Cannot locate repository_metadata with id {id}")
            return {}
        encoded_repository_id = trans.security.encode_id(repository_metadata.repository_id)
        repository_metadata_dict = repository_metadata.to_dict(
            view="element", value_mapper=self.__get_value_mapper(trans)
        )
        repository_metadata_dict["url"] = web.url_for(
            controller="repositories", action="show", id=encoded_repository_id
        )
        return repository_metadata_dict

    @web.legacy_expose_api
    def update(self, trans, payload, **kwd):
        """
        PUT /api/repository_revisions/{encoded_repository_metadata_id}/{payload}
        Updates the value of specified columns of the repository_metadata table based on the key / value pairs in payload.

        :param id: the encoded id of the `RepositoryMetadata` object
        """
        repository_metadata_id = kwd.get("id", None)
        if repository_metadata_id is None:
            raise HTTPBadRequest(detail="Missing required parameter 'id'.")
        repository_metadata = metadata_util.get_repository_metadata_by_id(trans.app, repository_metadata_id)
        if repository_metadata is None:
            decoded_repository_metadata_id = trans.security.decode_id(repository_metadata_id)
            log.debug(f"Cannot locate repository_metadata with id {decoded_repository_metadata_id}")
            return {}
        else:
            decoded_repository_metadata_id = repository_metadata.id
        flush_needed = False
        for key, new_value in payload.items():
            if hasattr(repository_metadata, key):
                # log information when setting attributes associated with the Tool Shed's install and test framework.
                if key in ["includes_tools", "missing_test_components"]:
                    log.debug(
                        "Setting repository_metadata column %s to value %s for changeset_revision %s via the Tool Shed API.",
                        key,
                        new_value,
                        repository_metadata.changeset_revision,
                    )
                setattr(repository_metadata, key, new_value)
                flush_needed = True
        if flush_needed:
            log.debug(
                "Updating repository_metadata record with id %s and changeset_revision %s.",
                decoded_repository_metadata_id,
                repository_metadata.changeset_revision,
            )
            trans.sa_session.add(repository_metadata)
            trans.sa_session.commit()
            trans.sa_session.refresh(repository_metadata)
        repository_metadata_dict = repository_metadata.to_dict(
            view="element", value_mapper=self.__get_value_mapper(trans)
        )
        repository_metadata_dict["url"] = web.url_for(
            controller="repository_revisions", action="show", id=repository_metadata_id
        )
        return repository_metadata_dict


def get_repository_metadata(session, downloadable, malicious, missing_test_components, includes_tools):
    stmt = select(RepositoryMetadata)
    if downloadable is not None:
        stmt = stmt.where(RepositoryMetadata.downloadable == util.asbool(downloadable))
    if malicious is not None:
        stmt = stmt.where(RepositoryMetadata.malicious == util.asbool(malicious))
    if missing_test_components is not None:
        stmt = stmt.where(RepositoryMetadata.missing_test_components == util.asbool(missing_test_components))
    if includes_tools is not None:
        stmt = stmt.where(RepositoryMetadata.includes_tools == util.asbool(includes_tools))
    stmt = stmt.order_by(RepositoryMetadata.repository_id.desc())
    return session.scalars(stmt)



import logging
import tempfile
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from sqlalchemy import (
    false,
    select,
)

from galaxy import util
from galaxy.tool_shed.metadata.metadata_generator import (
    BaseMetadataGenerator,
    HandleResultT,
    InvalidFileT,
)
from galaxy.util import inflector
from galaxy.web.form_builder import SelectField
from tool_shed.context import ProvidesRepositoriesContext
from tool_shed.repository_types import util as rt_util
from tool_shed.repository_types.metadata import TipOnly
from tool_shed.structured_app import ToolShedApp
from tool_shed.util import (
    basic_util,
    common_util,
    hg_util,
    metadata_util,
    repository_util,
    shed_util_common as suc,
    tool_util,
)
from tool_shed.util.metadata_util import repository_metadata_by_changeset_revision
from tool_shed.webapp.model import (
    Repository,
    RepositoryMetadata,
    User,
)
from tool_shed.webapp.model.db import get_repository_by_name_and_owner

log = logging.getLogger(__name__)


class ToolShedMetadataGenerator(BaseMetadataGenerator):
    """A MetadataGenerator building on ToolShed's app and repository constructs."""

    app: ToolShedApp
    repository: Optional[Repository]  # type:ignore[assignment]

    # why is mypy making me re-annotate these things from the base class, it didn't
    # when they were in the same file
    invalid_file_tups: List[InvalidFileT]
    repository_clone_url: Optional[str]

    def __init__(
        self,
        trans: ProvidesRepositoriesContext,
        repository: Optional[Repository] = None,
        changeset_revision: Optional[str] = None,
        repository_clone_url: Optional[str] = None,
        shed_config_dict: Optional[Dict[str, Any]] = None,
        relative_install_dir=None,
        repository_files_dir=None,
        resetting_all_metadata_on_repository=False,
        updating_installed_repository=False,
        persist=False,
        metadata_dict=None,
        user=None,
    ):
        self.trans = trans
        self.app = trans.app
        self.user = user
        self.repository = repository
        if changeset_revision is None and self.repository is not None:
            self.changeset_revision = self.repository.tip()
        else:
            self.changeset_revision = changeset_revision
        if repository_clone_url is None and self.repository is not None:
            self.repository_clone_url = common_util.generate_clone_url_for(self.trans, self.repository)
        else:
            self.repository_clone_url = repository_clone_url
        if shed_config_dict is None:
            self.shed_config_dict = {}
        else:
            self.shed_config_dict = shed_config_dict
        if relative_install_dir is None and self.repository is not None:
            relative_install_dir = self.repository.repo_path(self.app)
        if repository_files_dir is None and self.repository is not None:
            repository_files_dir = self.repository.repo_path(self.app)
        if metadata_dict is None:
            self.metadata_dict = {}
        else:
            self.metadata_dict = metadata_dict
        self.relative_install_dir = relative_install_dir
        self.repository_files_dir = repository_files_dir
        self.resetting_all_metadata_on_repository = resetting_all_metadata_on_repository
        self.updating_installed_repository = updating_installed_repository
        self.persist = persist
        self.invalid_file_tups = []
        self.sa_session = trans.app.model.session

    def initial_metadata_dict(self) -> Dict[str, Any]:
        return {}

    def set_repository(
        self, repository, relative_install_dir: Optional[str] = None, changeset_revision: Optional[str] = None
    ):
        self.repository = repository
        if relative_install_dir is None and self.repository is not None:
            relative_install_dir = repository.repo_path(self.app)
        if changeset_revision is None and self.repository is not None:
            self.set_changeset_revision(self.repository.tip())
        else:
            self.set_changeset_revision(changeset_revision)
        self.shed_config_dict = {}
        self._reset_attributes_after_repository_update(relative_install_dir)

    def handle_repository_elem(self, repository_elem, only_if_compiling_contained_td=False) -> HandleResultT:
        """
        Process the received repository_elem which is a <repository> tag either from a
        repository_dependencies.xml file or a tool_dependencies.xml file.  If the former,
        we're generating repository dependencies metadata for a repository in the Tool Shed.
        If the latter, we're generating package dependency metadata within Galaxy or the
        Tool Shed.
        """
        is_valid = True
        error_message = ""
        toolshed = repository_elem.get("toolshed", None)
        name = repository_elem.get("name", None)
        owner = repository_elem.get("owner", None)
        changeset_revision = repository_elem.get("changeset_revision", None)
        prior_installation_required = str(repository_elem.get("prior_installation_required", False))
        repository_dependency_tup = [
            toolshed,
            name,
            owner,
            changeset_revision,
            prior_installation_required,
            str(only_if_compiling_contained_td),
        ]
        if not toolshed:
            # Default to the current tool shed.
            toolshed = self.trans.repositories_hostname
            log.warning(f"\n\n\n\n\n\nin not toolshed with {toolshed}\n\n\n\n")
            # toolshed = str(url_for("/", qualified=True)).rstrip("/")
            repository_dependency_tup[0] = toolshed
        else:
            log.warning(f"moooocww.....{toolshed}\n\n\n\n\n")
        toolshed = common_util.remove_protocol_from_tool_shed_url(toolshed)

        if suc.tool_shed_is_this_tool_shed(toolshed, trans=self.trans):
            try:
                user = get_user_by_username(self.sa_session, owner)
            except Exception:
                error_message = (
                    f"Ignoring repository dependency definition for tool shed {toolshed}, name {name}, owner {owner}, "
                )
                error_message += f"changeset revision {changeset_revision} because the owner is invalid."
                log.debug(error_message)
                is_valid = False
                return repository_dependency_tup, is_valid, error_message
            try:
                repository = get_repository(self.sa_session, name, user.id)
            except Exception:
                error_message = f"Ignoring repository dependency definition for tool shed {toolshed},"
                error_message += f"name {name}, owner {owner}, "
                error_message += f"changeset revision {changeset_revision} because the name is invalid.  "
                log.debug(error_message)
                is_valid = False
                return repository_dependency_tup, is_valid, error_message
            repo = repository.hg_repo

            # The received changeset_revision may be None since defining it in the dependency definition is optional.
            # If this is the case, the default will be to set its value to the repository dependency tip revision.
            # This probably occurs only when handling circular dependency definitions.
            tip_ctx = repo[repo.changelog.tip()]
            # Make sure the repo.changlog includes at least 1 revision.
            if changeset_revision is None and tip_ctx.rev() >= 0:
                changeset_revision = str(tip_ctx)
                repository_dependency_tup = [
                    toolshed,
                    name,
                    owner,
                    changeset_revision,
                    prior_installation_required,
                    str(only_if_compiling_contained_td),
                ]
                return repository_dependency_tup, is_valid, error_message
            else:
                # Find the specified changeset revision in the repository's changelog to see if it's valid.
                found = False
                for changeset in repo.changelog:
                    changeset_hash = str(repo[changeset])
                    if changeset_hash == changeset_revision:
                        found = True
                        break
                if not found:
                    error_message = f"Ignoring repository dependency definition for tool shed {toolshed}, name {name}, owner {owner}, "
                    error_message += (
                        f"changeset revision {changeset_revision} because the changeset revision is invalid.  "
                    )
                    log.debug(error_message)
                    is_valid = False
                    return repository_dependency_tup, is_valid, error_message
        else:
            # Repository dependencies are currently supported within a single tool shed.
            error_message = "Repository dependencies are currently supported only within the same tool shed.  Ignoring "
            error_message += f"repository dependency definition  for tool shed {toolshed}, name {name}, owner {owner}, changeset revision {changeset_revision}.  "
            log.debug(error_message)
            is_valid = False
            return repository_dependency_tup, is_valid, error_message
        return repository_dependency_tup, is_valid, error_message


class RepositoryMetadataManager(ToolShedMetadataGenerator):
    def __init__(
        self,
        trans: ProvidesRepositoriesContext,
        repository=None,
        changeset_revision=None,
        repository_clone_url=None,
        shed_config_dict=None,
        relative_install_dir=None,
        repository_files_dir=None,
        resetting_all_metadata_on_repository=False,
        updating_installed_repository=False,
        persist=False,
        metadata_dict=None,
    ):
        super().__init__(
            trans,
            repository,
            changeset_revision,
            repository_clone_url,
            shed_config_dict,
            relative_install_dir,
            repository_files_dir,
            resetting_all_metadata_on_repository,
            updating_installed_repository,
            persist,
            metadata_dict=metadata_dict,
            user=trans.user,
        )
        app = trans.app
        user = trans.user
        self.sa_session = app.model.context
        self.app = app
        self.user = user
        # Repository metadata comparisons for changeset revisions.
        self.EQUAL = "equal"
        self.NO_METADATA = "no metadata"
        self.NOT_EQUAL_AND_NOT_SUBSET = "not equal and not subset"
        self.SUBSET = "subset"
        self.SUBSET_VALUES = [self.EQUAL, self.SUBSET]

    def _add_tool_versions(self, id: int, repository_metadata, changeset_revisions):
        # Build a dictionary of { 'tool id' : 'parent tool id' } pairs for each tool in repository_metadata.
        metadata = repository_metadata.metadata
        tool_versions_dict = {}
        for tool_dict in metadata.get("tools", []):
            # We have at least 2 changeset revisions to compare tool guids and tool ids.
            parent_id = self._get_parent_id(
                id, tool_dict["id"], tool_dict["version"], tool_dict["guid"], changeset_revisions
            )
            tool_versions_dict[tool_dict["guid"]] = parent_id
        if tool_versions_dict:
            repository_metadata.tool_versions = tool_versions_dict
            self.sa_session.add(repository_metadata)
            session = self.sa_session()
            session.commit()

    def build_repository_ids_select_field(
        self, name="repository_ids", multiple=True, display="checkboxes", my_writable=False
    ):
        """Generate the current list of repositories for resetting metadata."""
        repositories_select_field = SelectField(name=name, multiple=multiple, display=display)
        for repository in self.get_repositories_for_setting_metadata(my_writable=my_writable, order=True):
            owner = str(repository.user.username)
            option_label = f"{str(repository.name)} ({owner})"
            option_value = f"{self.app.security.encode_id(repository.id)}"
            repositories_select_field.add_option(option_label, option_value)
        return repositories_select_field

    def _clean_repository_metadata(self, changeset_revisions):
        assert self.repository
        # Delete all repository_metadata records associated with the repository that have
        # a changeset_revision that is not in changeset_revisions.  We sometimes see multiple
        # records with the same changeset revision value - no idea how this happens. We'll
        # assume we can delete the older records, so we'll order by update_time descending and
        # delete records that have the same changeset_revision we come across later.
        for repository_metadata in get_repository_metadata(self.sa_session, self.repository.id):
            changeset_revision = repository_metadata.changeset_revision
            if changeset_revision not in changeset_revisions:
                self.sa_session.delete(repository_metadata)
                session = self.sa_session()
                session.commit()

    def compare_changeset_revisions(self, ancestor_changeset_revision, ancestor_metadata_dict):
        """
        Compare the contents of two changeset revisions to determine if a new repository
        metadata revision should be created.
        """
        # The metadata associated with ancestor_changeset_revision is ancestor_metadata_dict.
        # This changeset_revision is an ancestor of self.changeset_revision which is associated
        # with self.metadata_dict.  A new repository_metadata record will be created only
        # when this method returns the constant value self.NOT_EQUAL_AND_NOT_SUBSET.
        ancestor_tools = ancestor_metadata_dict.get("tools", [])
        ancestor_guids = [tool_dict["guid"] for tool_dict in ancestor_tools]
        ancestor_guids.sort()
        ancestor_repository_dependencies_dict = ancestor_metadata_dict.get("repository_dependencies", {})
        ancestor_repository_dependencies = ancestor_repository_dependencies_dict.get("repository_dependencies", [])
        ancestor_tool_dependencies = ancestor_metadata_dict.get("tool_dependencies", {})
        ancestor_data_manager = ancestor_metadata_dict.get("data_manager", {})
        current_tools = self.metadata_dict.get("tools", [])
        current_guids = [tool_dict["guid"] for tool_dict in current_tools]
        current_guids.sort()
        current_repository_dependencies_dict = self.metadata_dict.get("repository_dependencies", {})
        current_repository_dependencies = current_repository_dependencies_dict.get("repository_dependencies", [])
        current_tool_dependencies = self.metadata_dict.get("tool_dependencies", {})
        current_data_manager = self.metadata_dict.get("data_manager", {})
        # Handle case where no metadata exists for either changeset.
        no_repository_dependencies = not ancestor_repository_dependencies and not current_repository_dependencies
        no_tool_dependencies = not ancestor_tool_dependencies and not current_tool_dependencies
        no_tools = not ancestor_guids and not current_guids
        no_data_manager = not ancestor_data_manager and not current_data_manager
        if no_repository_dependencies and no_tool_dependencies and no_tools and no_data_manager:
            return self.NO_METADATA
        repository_dependency_comparison = self.compare_repository_dependencies(
            ancestor_repository_dependencies, current_repository_dependencies
        )
        tool_dependency_comparison = self.compare_tool_dependencies(
            ancestor_tool_dependencies, current_tool_dependencies
        )
        data_manager_comparison = self.compare_data_manager(ancestor_data_manager, current_data_manager)
        # Handle case where all metadata is the same.
        if (
            ancestor_guids == current_guids
            and repository_dependency_comparison == self.EQUAL
            and tool_dependency_comparison == self.EQUAL
            and data_manager_comparison == self.EQUAL
        ):
            return self.EQUAL
        # Handle case where ancestor metadata is a subset of current metadata.
        # readme_file_is_subset = readme_file_comparision in [ self.EQUAL, self.SUBSET ]
        repository_dependency_is_subset = repository_dependency_comparison in self.SUBSET_VALUES
        tool_dependency_is_subset = tool_dependency_comparison in self.SUBSET_VALUES
        datamanager_is_subset = data_manager_comparison in self.SUBSET_VALUES
        if repository_dependency_is_subset and tool_dependency_is_subset and datamanager_is_subset:
            is_subset = True
            for guid in ancestor_guids:
                if guid not in current_guids:
                    is_subset = False
                    break
            if is_subset:
                return self.SUBSET
        return self.NOT_EQUAL_AND_NOT_SUBSET

    def compare_data_manager(self, ancestor_metadata, current_metadata):
        """Determine if ancestor_metadata is the same as or a subset of current_metadata for data_managers."""

        def __data_manager_dict_to_tuple_list(metadata_dict):
            # we do not check tool_guid or tool conf file name
            return {
                (
                    name,
                    tuple(sorted(value.get("data_tables", []))),
                    value.get("guid"),
                    value.get("version"),
                    value.get("name"),
                    value.get("id"),
                )
                for name, value in metadata_dict.items()
            }

        # only compare valid entries, any invalid entries are ignored
        ancestor_metadata = __data_manager_dict_to_tuple_list(ancestor_metadata.get("data_managers", {}))
        current_metadata = __data_manager_dict_to_tuple_list(current_metadata.get("data_managers", {}))
        # use set comparisons
        if ancestor_metadata.issubset(current_metadata):
            if ancestor_metadata == current_metadata:
                return self.EQUAL
            return self.SUBSET
        return self.NOT_EQUAL_AND_NOT_SUBSET

    def compare_repository_dependencies(self, ancestor_repository_dependencies, current_repository_dependencies):
        """
        Determine if ancestor_repository_dependencies is the same as or a subset of
        current_repository_dependencies.
        """
        # The list of repository_dependencies looks something like:
        # [["http://localhost:9009", "emboss_datatypes", "test", "ab03a2a5f407", "False", "False"]].
        # Create a string from each tuple in the list for easier comparison.
        if len(ancestor_repository_dependencies) <= len(current_repository_dependencies):
            for ancestor_tup in ancestor_repository_dependencies:
                (
                    a_tool_shed,
                    a_repo_name,
                    a_repo_owner,
                    a_changeset_revision,
                    a_prior_installation_required,
                    a_only_if_compiling_contained_td,
                ) = ancestor_tup
                cleaned_a_tool_shed = common_util.remove_protocol_from_tool_shed_url(a_tool_shed)
                found_in_current = False
                for current_tup in current_repository_dependencies:
                    (
                        c_tool_shed,
                        c_repo_name,
                        c_repo_owner,
                        c_changeset_revision,
                        c_prior_installation_required,
                        c_only_if_compiling_contained_td,
                    ) = current_tup
                    cleaned_c_tool_shed = common_util.remove_protocol_from_tool_shed_url(c_tool_shed)
                    if (
                        cleaned_c_tool_shed == cleaned_a_tool_shed
                        and c_repo_name == a_repo_name
                        and c_repo_owner == a_repo_owner
                        and c_changeset_revision == a_changeset_revision
                        and util.string_as_bool(c_prior_installation_required)
                        == util.string_as_bool(a_prior_installation_required)
                        and util.string_as_bool(c_only_if_compiling_contained_td)
                        == util.string_as_bool(a_only_if_compiling_contained_td)
                    ):
                        found_in_current = True
                        break
                if not found_in_current:
                    # In some cases, the only difference between a dependency definition in the lists
                    # is the changeset_revision value.  We'll check to see if this is the case, and if
                    # the defined dependency is a repository that has metadata set only on its tip.
                    if not self.different_revision_defines_tip_only_repository_dependency(
                        ancestor_tup, current_repository_dependencies
                    ):
                        return self.NOT_EQUAL_AND_NOT_SUBSET
                    return self.SUBSET
            if len(ancestor_repository_dependencies) == len(current_repository_dependencies):
                return self.EQUAL
            else:
                return self.SUBSET
        return self.NOT_EQUAL_AND_NOT_SUBSET

    def compare_tool_dependencies(self, ancestor_tool_dependencies, current_tool_dependencies):
        """
        Determine if ancestor_tool_dependencies is the same as or a subset of current_tool_dependencies.
        """
        # The tool_dependencies dictionary looks something like:
        # {'bwa/0.5.9': {'readme': 'some string', 'version': '0.5.9', 'type': 'package', 'name': 'bwa'}}
        if len(ancestor_tool_dependencies) <= len(current_tool_dependencies):
            for ancestor_td_key in ancestor_tool_dependencies.keys():
                if ancestor_td_key in current_tool_dependencies:
                    # The only values that could have changed between the 2 dictionaries are the
                    # "readme" or "type" values.  Changing the readme value makes no difference.
                    # Changing the type will change the installation process, but for now we'll
                    # assume it was a typo, so new metadata shouldn't be generated.
                    continue
                else:
                    return self.NOT_EQUAL_AND_NOT_SUBSET
            # At this point we know that ancestor_tool_dependencies is at least a subset of current_tool_dependencies.
            if len(ancestor_tool_dependencies) == len(current_tool_dependencies):
                return self.EQUAL
            else:
                return self.SUBSET
        return self.NOT_EQUAL_AND_NOT_SUBSET

    def create_or_update_repository_metadata(self, changeset_revision, metadata_dict):
        """Create or update a repository_metadata record in the tool shed."""
        has_repository_dependencies = False
        has_repository_dependencies_only_if_compiling_contained_td = False
        includes_tools = False
        includes_tool_dependencies = False
        if metadata_dict:
            repository_dependencies_dict = metadata_dict.get("repository_dependencies", {})
            repository_dependencies = repository_dependencies_dict.get("repository_dependencies", [])
            (
                has_repository_dependencies,
                has_repository_dependencies_only_if_compiling_contained_td,
            ) = repository_util.get_repository_dependency_types(repository_dependencies)
            if "tools" in metadata_dict:
                includes_tools = True
            if "tool_dependencies" in metadata_dict:
                includes_tool_dependencies = True
        if (
            has_repository_dependencies
            or has_repository_dependencies_only_if_compiling_contained_td
            or includes_tools
            or includes_tool_dependencies
        ):
            downloadable = True
        else:
            downloadable = False
        assert self.repository
        repository_metadata = repository_metadata_by_changeset_revision(
            self.app.model, self.repository.id, changeset_revision
        )
        if repository_metadata:
            repository_metadata.metadata = metadata_dict
            repository_metadata.downloadable = downloadable
            repository_metadata.has_repository_dependencies = has_repository_dependencies
            repository_metadata.includes_datatypes = False
            repository_metadata.includes_tools = includes_tools
            repository_metadata.includes_tool_dependencies = includes_tool_dependencies
            repository_metadata.includes_workflows = False
        else:
            repository_metadata = RepositoryMetadata(
                repository_id=self.repository.id,
                changeset_revision=changeset_revision,
                metadata=metadata_dict,
                downloadable=downloadable,
                has_repository_dependencies=has_repository_dependencies,
                includes_datatypes=False,
                includes_tools=includes_tools,
                includes_tool_dependencies=includes_tool_dependencies,
                includes_workflows=False,
            )
        assert repository_metadata
        # Always set the default values for the following columns.  When resetting all metadata
        # on a repository this will reset the values.
        assert repository_metadata
        repository_metadata.missing_test_components = False
        self.sa_session.add(repository_metadata)
        session = self.sa_session()
        session.commit()

        return repository_metadata

    def different_revision_defines_tip_only_repository_dependency(self, rd_tup, repository_dependencies):
        """
        Determine if the only difference between rd_tup and a dependency definition in the list of
        repository_dependencies is the changeset_revision value.
        """
        (
            rd_tool_shed,
            rd_name,
            rd_owner,
            rd_changeset_revision,
            rd_prior_installation_required,
            rd_only_if_compiling_contained_td,
        ) = common_util.parse_repository_dependency_tuple(rd_tup)
        cleaned_rd_tool_shed = common_util.remove_protocol_from_tool_shed_url(rd_tool_shed)
        for repository_dependency in repository_dependencies:
            (
                tool_shed,
                name,
                owner,
                changeset_revision,
                prior_installation_required,
                only_if_compiling_contained_td,
            ) = common_util.parse_repository_dependency_tuple(repository_dependency)
            cleaned_tool_shed = common_util.remove_protocol_from_tool_shed_url(tool_shed)
            if cleaned_rd_tool_shed == cleaned_tool_shed and rd_name == name and rd_owner == owner:
                # Determine if the repository represented by the dependency tuple is an instance of the repository type TipOnly.
                required_repository = get_repository_by_name_and_owner(self.app.model.context, name, owner)
                repository_type_class = self.app.repository_types_registry.get_class_by_label(required_repository.type)
                return isinstance(repository_type_class, TipOnly)
        return False

    def _get_parent_id(self, id: int, old_id, version, guid, changeset_revisions):
        parent_id = None
        # Compare from most recent to oldest.
        changeset_revisions.reverse()
        for changeset_revision in changeset_revisions:
            repository_metadata = repository_metadata_by_changeset_revision(self.app.model, id, changeset_revision)
            assert repository_metadata
            metadata = repository_metadata.metadata
            tools_dicts = metadata.get("tools", [])
            for tool_dict in tools_dicts:
                if tool_dict["guid"] == guid:
                    # The tool has not changed between the compared changeset revisions.
                    continue
                if tool_dict["id"] == old_id and tool_dict["version"] != version:
                    # The tool version is different, so we've found the parent.
                    return tool_dict["guid"]
        if parent_id is None:
            # The tool did not change through all of the changeset revisions.
            return old_id

    def get_repositories_for_setting_metadata(self, my_writable=False, order=True):
        """
        Return a list of repositories for resetting metadata.  The order parameter
        is used for displaying the list of repositories ordered alphabetically for display on
        a page.  When called from the Tool Shed API, order is False.
        """
        # When called from the Tool Shed API, the metadata is reset on all repositories of types
        # repository_suite_definition and tool_dependency_definition in addition to other selected
        # repositories.
        if my_writable:
            username = self.user.username
            repo_ids = []
            for repository in get_current_repositories(self.sa_session):
                # Always reset metadata on all repositories of types repository_suite_definition and
                # tool_dependency_definition.
                if repository.type in [rt_util.REPOSITORY_SUITE_DEFINITION, rt_util.TOOL_DEPENDENCY_DEFINITION]:
                    repo_ids.append(repository.id)
                else:
                    allow_push = repository.allow_push()
                    if allow_push:
                        # Include all repositories that are writable by the current user.
                        allow_push_usernames = allow_push.split(",")
                        if username in allow_push_usernames:
                            repo_ids.append(repository.id)
            if repo_ids:
                return get_filtered_repositories(self.sa_session, repo_ids, order)
            else:
                return []
        else:
            return get_current_repositories(self.sa_session, order)

    def new_metadata_required_for_utilities(self):
        """
        This method compares the last stored repository_metadata record associated with self.repository
        against the contents of self.metadata_dict and returns True or False for the union set of Galaxy
        utilities contained in both metadata dictionaries.  The metadata contained in self.metadata_dict
        may not be a subset of that contained in the last stored repository_metadata record associated with
        self.repository because one or more Galaxy utilities may have been deleted from self.repository in
        the new tip.
        """
        assert self.repository
        repository_metadata = metadata_util.get_latest_repository_metadata(
            self.app, self.repository.id, downloadable=False
        )
        repository_dependencies_required = self.new_repository_dependency_metadata_required(repository_metadata)
        tools_required = self.new_tool_metadata_required(repository_metadata)
        tool_dependencies_required = self.new_tool_dependency_metadata_required(repository_metadata)
        data_managers_required = self.new_data_manager_required(repository_metadata)
        if repository_dependencies_required or tools_required or tool_dependencies_required or data_managers_required:
            return True
        return False

    def new_repository_dependency_metadata_required(self, repository_metadata):
        """
        Compare the last saved metadata for each repository dependency in the repository
        with the new metadata in self.metadata_dict to determine if a new repository_metadata
        table record is required or if the last saved metadata record can be updated for
        repository_dependencies instead.
        """
        if repository_metadata:
            metadata = repository_metadata.metadata
            if "repository_dependencies" in metadata:
                saved_repository_dependencies = metadata["repository_dependencies"]["repository_dependencies"]
                new_repository_dependencies_metadata = self.metadata_dict.get("repository_dependencies", None)
                if new_repository_dependencies_metadata:
                    new_repository_dependencies = self.metadata_dict["repository_dependencies"][
                        "repository_dependencies"
                    ]
                    # TODO: We used to include the following here to handle the case where repository
                    # dependency definitions were deleted.  However this erroneously returned True in
                    # cases where is should not have done so.  This usually occurred where multiple single
                    # files were uploaded when a single tarball should have been.  We need to implement
                    # support for handling deleted repository dependency definitions so that we can guarantee
                    # reproducibility, but we need to do it in a way that is better than the following.
                    # for new_repository_dependency in new_repository_dependencies:
                    #     if new_repository_dependency not in saved_repository_dependencies:
                    #         return True
                    # The saved metadata must be a subset of the new metadata.
                    for saved_repository_dependency in saved_repository_dependencies:
                        if saved_repository_dependency not in new_repository_dependencies:
                            # In some cases, the only difference between a dependency definition in the lists
                            # is the changeset_revision value.  We'll check to see if this is the case, and if
                            # the defined dependency is a repository that has metadata set only on its tip.
                            if not self.different_revision_defines_tip_only_repository_dependency(
                                saved_repository_dependency, new_repository_dependencies
                            ):
                                return True
                    return False
                else:
                    # The repository_dependencies.xml file must have been deleted, so create a new
                    # repository_metadata record so we always have access to the deleted file.
                    return True
            else:
                return False
        else:
            if "repository_dependencies" in self.metadata_dict:
                # There is no saved repository metadata, so we need to create a new repository_metadata record.
                return True
            else:
                # self.metadata_dict includes no metadata for repository dependencies, so a new repository_metadata
                # record is not needed.
                return False

    def new_data_manager_required(self, repository_metadata):
        if self.metadata_dict and repository_metadata and repository_metadata.metadata:
            return self.compare_data_manager(self.metadata_dict, repository_metadata.metadata) != self.EQUAL
        else:
            return bool(
                repository_metadata
                and repository_metadata.metadata
                and repository_metadata.metadata.get("data_managers")
            )

    def new_tool_metadata_required(self, repository_metadata):
        """
        Compare the last saved metadata for each tool in the repository with the new metadata in
        self.metadata_dict to determine if a new repository_metadata table record is required, or if
        the last saved metadata record can be updated instead.
        """
        if "tools" in self.metadata_dict:
            if repository_metadata:
                metadata = repository_metadata.metadata
                if metadata:
                    if "tools" in metadata:
                        saved_tool_ids = []
                        # The metadata for one or more tools was successfully generated in the past
                        # for this repository, so we first compare the version string for each tool id
                        # in self.metadata_dict with what was previously saved to see if we need to create
                        # a new table record or if we can simply update the existing record.
                        for new_tool_metadata_dict in self.metadata_dict["tools"]:
                            for saved_tool_metadata_dict in metadata["tools"]:
                                if saved_tool_metadata_dict["id"] not in saved_tool_ids:
                                    saved_tool_ids.append(saved_tool_metadata_dict["id"])
                                if new_tool_metadata_dict["id"] == saved_tool_metadata_dict["id"]:
                                    if new_tool_metadata_dict["version"] != saved_tool_metadata_dict["version"]:
                                        return True
                        # So far, a new metadata record is not required, but we still have to check to see if
                        # any new tool ids exist in self.metadata_dict that are not in the saved metadata.  We do
                        # this because if a new tarball was uploaded to a repository that included tools, it
                        # may have removed existing tool files if they were not included in the uploaded tarball.
                        for new_tool_metadata_dict in self.metadata_dict["tools"]:
                            if new_tool_metadata_dict["id"] not in saved_tool_ids:
                                return True
                        return False
                    else:
                        # The new metadata includes tools, but the stored metadata does not, so we can
                        # update the stored metadata.
                        return False
                else:
                    # There is no stored metadata, so we can update the metadata column in the
                    # repository_metadata table.
                    return False
            else:
                # There is no stored repository metadata, so we need to create a new repository_metadata
                # table record.
                return True
        # self.metadata_dict includes no metadata for tools, so a new repository_metadata table
        # record is not needed.
        return False

    def new_tool_dependency_metadata_required(self, repository_metadata):
        """
        Compare the last saved metadata for each tool dependency in the repository with the new
        metadata in self.metadata_dict to determine if a new repository_metadata table record is
        required or if the last saved metadata record can be updated for tool_dependencies instead.
        """
        if repository_metadata:
            metadata = repository_metadata.metadata
            if metadata:
                if "tool_dependencies" in metadata:
                    saved_tool_dependencies = metadata["tool_dependencies"]
                    new_tool_dependencies = self.metadata_dict.get("tool_dependencies", None)
                    if new_tool_dependencies:
                        # TODO: We used to include the following here to handle the case where
                        # tool dependency definitions were deleted.  However, this erroneously
                        # returned True in cases where is should not have done so.  This usually
                        # occurred where multiple single files were uploaded when a single tarball
                        # should have been.  We need to implement support for handling deleted
                        # tool dependency definitions so that we can guarantee reproducibility,
                        # but we need to do it in a way that is better than the following.
                        # for new_tool_dependency in new_tool_dependencies:
                        #     if new_tool_dependency not in saved_tool_dependencies:
                        #         return True
                        # The saved metadata must be a subset of the new metadata.
                        for saved_tool_dependency in saved_tool_dependencies:
                            if saved_tool_dependency not in new_tool_dependencies:
                                return True
                        return False
                    else:
                        # The tool_dependencies.xml file must have been deleted, so create a new
                        # repository_metadata record so we always have
                        # access to the deleted file.
                        return True
                else:
                    return False
            else:
                # We have repository metadata that does not include metadata for any tool dependencies
                # in the repository, so we can update the existing repository metadata.
                return False
        else:
            if "tool_dependencies" in self.metadata_dict:
                # There is no saved repository metadata, so we need to create a new repository_metadata
                # record.
                return True
            else:
                # self.metadata_dict includes no metadata for tool dependencies, so a new repository_metadata
                # record is not needed.
                return False

    def reset_all_metadata_on_repository_in_tool_shed(self, repository_clone_url=None):
        """Reset all metadata on a single repository in a tool shed."""
        assert self.repository
        log.debug(f"Resetting all metadata on repository: {self.repository.name}")
        repo = self.repository.hg_repo
        # The list of changeset_revisions refers to repository_metadata records that have been created
        # or updated.  When the following loop completes, we'll delete all repository_metadata records
        # for this repository that do not have a changeset_revision value in this list.
        changeset_revisions: List[Optional[str]] = []
        # When a new repository_metadata record is created, it always uses the values of
        # metadata_changeset_revision and metadata_dict.
        metadata_changeset_revision = None
        metadata_dict = None
        ancestor_changeset_revision = None
        ancestor_metadata_dict = None
        for changeset in self.repository.get_changesets_for_setting_metadata(self.app):
            work_dir = tempfile.mkdtemp(prefix="tmp-toolshed-ramorits")
            ctx = repo[changeset]
            log.debug("Cloning repository changeset revision: %s", str(ctx.rev()))
            assert self.repository_clone_url
            repository_clone_url = repository_clone_url or self.repository_clone_url
            cloned_ok, error_message = hg_util.clone_repository(repository_clone_url, work_dir, str(ctx.rev()))
            if cloned_ok:
                log.debug("Generating metadata for changeset revision: %s", str(ctx.rev()))
                self.set_changeset_revision(str(ctx))
                self.set_repository_files_dir(work_dir)
                self.generate_metadata_for_changeset_revision()
                if self.metadata_dict:
                    if metadata_changeset_revision is None and metadata_dict is None:
                        # We're at the first change set in the change log.
                        metadata_changeset_revision = self.changeset_revision
                        metadata_dict = self.metadata_dict
                    if ancestor_changeset_revision:
                        # Compare metadata from ancestor and current.  The value of comparison will be one of:
                        # self.NO_METADATA - no metadata for either ancestor or current, so continue from current
                        # self.EQUAL - ancestor metadata is equivalent to current metadata, so continue from current
                        # self.SUBSET - ancestor metadata is a subset of current metadata, so continue from current
                        # self.NOT_EQUAL_AND_NOT_SUBSET - ancestor metadata is neither equal to nor a subset of current
                        # metadata, so persist ancestor metadata.
                        log.info(f"amd {ancestor_metadata_dict}")
                        comparison = self.compare_changeset_revisions(
                            ancestor_changeset_revision, ancestor_metadata_dict
                        )
                        log.info(f"comparison {comparison}")
                        if comparison in [self.NO_METADATA, self.EQUAL, self.SUBSET]:
                            ancestor_changeset_revision = self.changeset_revision
                            ancestor_metadata_dict = self.metadata_dict
                        elif comparison == self.NOT_EQUAL_AND_NOT_SUBSET:
                            metadata_changeset_revision = ancestor_changeset_revision
                            metadata_dict = ancestor_metadata_dict
                            self.create_or_update_repository_metadata(metadata_changeset_revision, metadata_dict)
                            changeset_revisions.append(metadata_changeset_revision)
                            ancestor_changeset_revision = self.changeset_revision
                            ancestor_metadata_dict = self.metadata_dict
                    else:
                        # We're at the beginning of the change log.
                        ancestor_changeset_revision = self.changeset_revision
                        ancestor_metadata_dict = self.metadata_dict
                    if not ctx.children():
                        metadata_changeset_revision = self.changeset_revision
                        metadata_dict = self.metadata_dict
                        # We're at the end of the change log.
                        self.create_or_update_repository_metadata(metadata_changeset_revision, metadata_dict)
                        changeset_revisions.append(metadata_changeset_revision)
                        ancestor_changeset_revision = None
                        ancestor_metadata_dict = None
                elif ancestor_metadata_dict:
                    # We reach here only if self.metadata_dict is empty and ancestor_metadata_dict is not.
                    if not ctx.children():
                        # We're at the end of the change log.
                        self.create_or_update_repository_metadata(metadata_changeset_revision, metadata_dict)
                        changeset_revisions.append(metadata_changeset_revision)
                        ancestor_changeset_revision = None
                        ancestor_metadata_dict = None
            basic_util.remove_dir(work_dir)
        # Delete all repository_metadata records for this repository that do not have a changeset_revision
        # value in changeset_revisions.
        self._clean_repository_metadata(changeset_revisions)
        # Set tool version information for all downloadable changeset revisions.  Get the list of changeset
        # revisions from the changelog.
        self._reset_all_tool_versions(repo)

    def _reset_all_tool_versions(self, repo):
        """Reset tool version lineage for those changeset revisions that include valid tools."""
        assert self.repository
        changeset_revisions_that_contain_tools = _get_changeset_revisions_that_contain_tools(
            self.app, repo, self.repository
        )
        # The list of changeset_revisions_that_contain_tools is now filtered to contain only those that
        # are downloadable and contain tools.  If a repository includes tools, build a dictionary of
        # { 'tool id' : 'parent tool id' } pairs for each tool in each changeset revision.
        for index, changeset_revision in enumerate(changeset_revisions_that_contain_tools):
            tool_versions_dict = {}
            repository_metadata = repository_metadata_by_changeset_revision(
                self.app.model, self.repository.id, changeset_revision
            )
            assert repository_metadata
            metadata = repository_metadata.metadata
            tool_dicts = metadata["tools"]
            if index == 0:
                # The first changeset_revision is a special case because it will have no ancestor
                # changeset_revisions in which to match tools.  The parent tool id for tools in the
                # first changeset_revision will be the "old_id" in the tool config.
                for tool_dict in tool_dicts:
                    tool_versions_dict[tool_dict["guid"]] = tool_dict["id"]
            else:
                for tool_dict in tool_dicts:
                    parent_id = self._get_parent_id(
                        self.repository.id,
                        tool_dict["id"],
                        tool_dict["version"],
                        tool_dict["guid"],
                        changeset_revisions_that_contain_tools[0:index],
                    )
                    tool_versions_dict[tool_dict["guid"]] = parent_id
            if tool_versions_dict:
                repository_metadata.tool_versions = tool_versions_dict
                self.sa_session.add(repository_metadata)
                session = self.sa_session()
                session.commit()

    def reset_metadata_on_selected_repositories(self, **kwd):
        """
        Inspect the repository changelog to reset metadata for all appropriate changeset revisions.
        This method is called from both Galaxy and the Tool Shed.
        """
        message = ""
        status = "done"
        if repository_ids := util.listify(kwd.get("repository_ids", None)):
            successful_count = 0
            unsuccessful_count = 0
            for repository_id in repository_ids:
                try:
                    repository = repository_util.get_repository_in_tool_shed(self.app, repository_id)
                    self.set_repository(repository)
                    self.resetting_all_metadata_on_repository = True
                    self.reset_all_metadata_on_repository_in_tool_shed()
                    if self.invalid_file_tups:
                        message = tool_util.generate_message_for_invalid_tools(
                            self.app, self.invalid_file_tups, repository, None, as_html=False
                        )
                        log.debug(message)
                        unsuccessful_count += 1
                    else:
                        log.debug(
                            "Successfully reset metadata on repository %s owned by %s",
                            repository.name,
                            repository.user.username,
                        )
                        successful_count += 1
                except Exception:
                    log.exception("Error attempting to reset metadata on repository %s", str(repository.name))
                    unsuccessful_count += 1
            message = "Successfully reset metadata on {} {}.  ".format(
                successful_count,
                inflector.cond_plural(successful_count, "repository"),
            )
            if unsuccessful_count:
                message += "Error setting metadata on {} {} - see the paster log for details.  ".format(
                    unsuccessful_count,
                    inflector.cond_plural(unsuccessful_count, "repository"),
                )
        else:
            message = "Select at least one repository to on which to reset all metadata."
            status = "error"
        return message, status

    def set_repository(
        self, repository, relative_install_dir: Optional[str] = None, changeset_revision: Optional[str] = None
    ):
        super().set_repository(repository)
        self.repository_clone_url = relative_install_dir or common_util.generate_clone_url_for(self.trans, repository)

    def set_repository_metadata(self, host, content_alert_str="", **kwd):
        """
        Set metadata using the self.repository's current disk files, returning specific error
        messages (if any) to alert the repository owner that the changeset has problems.
        """
        assert self.repository
        message = ""
        status = "done"
        repository_id = self.repository.id
        repo = self.repository.hg_repo
        self.generate_metadata_for_changeset_revision()
        if self.metadata_dict:
            repository_metadata = None
            repository_type_class = self.app.repository_types_registry.get_class_by_label(self.repository.type)
            tip_only = isinstance(repository_type_class, TipOnly)
            if not tip_only and self.new_metadata_required_for_utilities():
                # Create a new repository_metadata table row.
                repository_metadata = self.create_or_update_repository_metadata(
                    self.repository.tip(), self.metadata_dict
                )
                # If this is the first record stored for this repository, see if we need to send any email alerts.
                if len(self.repository.downloadable_revisions) == 1:
                    suc.handle_email_alerts(
                        self.app, host, self.repository, content_alert_str="", new_repo_alert=True, admin_only=False
                    )
            else:
                # Update the latest stored repository metadata with the contents and attributes of self.metadata_dict.
                repository_metadata = metadata_util.get_latest_repository_metadata(
                    self.app, repository_id, downloadable=False
                )
                if repository_metadata:
                    downloadable = metadata_util.is_downloadable(self.metadata_dict)
                    # Update the last saved repository_metadata table row.
                    repository_metadata.changeset_revision = self.repository.tip()
                    repository_metadata.metadata = self.metadata_dict
                    repository_metadata.downloadable = downloadable
                    repository_metadata.includes_datatypes = False
                    # We don't store information about the special type of repository dependency that is needed only for
                    # compiling a tool dependency defined for the dependent repository.
                    repository_dependencies_dict = self.metadata_dict.get("repository_dependencies", {})
                    repository_dependencies = repository_dependencies_dict.get("repository_dependencies", [])
                    (
                        has_repository_dependencies,
                        has_repository_dependencies_only_if_compiling_contained_td,
                    ) = repository_util.get_repository_dependency_types(repository_dependencies)
                    repository_metadata.has_repository_dependencies = has_repository_dependencies
                    if "tool_dependencies" in self.metadata_dict:
                        repository_metadata.includes_tool_dependencies = True
                    else:
                        repository_metadata.includes_tool_dependencies = False
                    if "tools" in self.metadata_dict:
                        repository_metadata.includes_tools = True
                    else:
                        repository_metadata.includes_tools = False
                    repository_metadata.includes_workflows = False
                    repository_metadata.missing_test_components = False
                    self.sa_session.add(repository_metadata)
                    session = self.sa_session()
                    session.commit()
                else:
                    # There are no metadata records associated with the repository.
                    repository_metadata = self.create_or_update_repository_metadata(
                        self.repository.tip(), self.metadata_dict
                    )
            if "tools" in self.metadata_dict and repository_metadata and status != "error":
                # Set tool versions on the new downloadable change set.  The order of the list of changesets is
                # critical, so we use the repo's changelog.
                changeset_revisions = []
                for changeset in repo.changelog:
                    changeset_revision = str(repo[changeset])
                    if repository_metadata_by_changeset_revision(self.app.model, repository_id, changeset_revision):
                        changeset_revisions.append(changeset_revision)
                self._add_tool_versions(repository_id, repository_metadata, changeset_revisions)
        elif len(repo) == 1 and not self.invalid_file_tups:
            message = (
                f"Revision <b>{self.repository.tip()}</b> includes no Galaxy utilities for which metadata can "
                "be defined so this revision cannot be automatically installed into a local Galaxy instance."
            )
            status = "error"
        if self.invalid_file_tups:
            message = tool_util.generate_message_for_invalid_tools(
                self.app, self.invalid_file_tups, self.repository, self.metadata_dict
            )
            status = "error"
        return message, status

    def set_repository_metadata_due_to_new_tip(self, host, content_alert_str=None, **kwd):
        """Set metadata on the tip of self.repository in the tool shed."""
        error_message, status = self.set_repository_metadata(host, content_alert_str=content_alert_str, **kwd)
        return status, error_message


def _get_changeset_revisions_that_contain_tools(app: "ToolShedApp", repo, repository) -> List[str]:
    changeset_revisions_that_contain_tools = []
    for changeset in repo.changelog:
        changeset_revision = str(repo[changeset])
        repository_metadata = repository_metadata_by_changeset_revision(app.model, repository.id, changeset_revision)
        if repository_metadata:
            metadata = repository_metadata.metadata
            if metadata:
                if metadata.get("tools", None):
                    changeset_revisions_that_contain_tools.append(changeset_revision)
    return changeset_revisions_that_contain_tools


def get_user_by_username(session, username):
    stmt = select(User).where(User.username == username)
    return session.execute(stmt).scalar_one()


def get_repository(session, name, user_id):
    stmt = select(Repository).where(Repository.name == name).where(Repository.user_id == user_id)
    return session.execute(stmt).scalar_one()


def get_repository_metadata(session, repository_id):
    stmt = (
        select(RepositoryMetadata)
        .where(RepositoryMetadata.repository_id == repository_id)
        .order_by(RepositoryMetadata.changeset_revision, RepositoryMetadata.update_time.desc())  # type: ignore[attr-defined]  # mapped attribute
    )
    return session.scalars(stmt)


def get_current_repositories(session, order=False):
    stmt = select(Repository).where(Repository.deleted == false())
    if order:
        stmt = stmt.order_by(Repository.name, Repository.user_id)
    return session.scalars(stmt)


def get_filtered_repositories(session, repo_ids, order):
    stmt = select(Repository).where(Repository.id.in_(repo_ids))
    if order:
        stmt = stmt.order_by(Repository.name, Repository.user_id)
    return session.scalars(stmt)
